const CORNER_TOP_LEFT = one(UInt8)<<0
const CORNER_TOP_RIGHT = one(UInt8)<<1
const CORNER_BOTTOM_LEFT = one(UInt8)<<2
const CORNER_BOTTOM_RIGHT = one(UInt8)<<3


"""
Returns true if the given tile is a corner.
Note that we require □ be within the game border.
"""
function is_a_corner(game::Game, □::TileIndex)
    for dir in DIRECTIONS
        if ((game.board[□ + game.step_fore[dir]] & WALL) > 0) &&
           ((game.board[□ + game.step_left[dir]] & WALL) > 0)
            return true
        end
    end
    return false
end

struct Corner
    □::TileIndex
    types::UInt8 # flags - see CORNER_TOP_LEFT, etc. above
    edge_lengths::MVector{N_DIRS, Int32}
end

function find_corners(game::Game)
    # Finds corners, including interior ones
    # 01              Non-goal corner
    #                 ===============
    #  #               A non-goal corner square is an illegal square.
    # #-
    #

    corners = Corner[]
    deadlock = false
    for □ in tile_indices(game)
        # Check if □ is a corner
        if is_tile_within_border(game, □)
            v = game.board[□]
            wall_up = (game.board[□+game.step_fore[DIR_UP]] & WALL) > 0
            wall_down = (game.board[□+game.step_fore[DIR_DOWN]] & WALL) > 0
            wall_left = (game.board[□+game.step_fore[DIR_LEFT]] & WALL) > 0
            wall_right = (game.board[□+game.step_fore[DIR_RIGHT]] & WALL) > 0
            if ((v & WALL) == 0) && # not a wall
               (wall_up || wall_down) && # vertical wall neighbor
               (wall_right || wall_left) # horizontal wall neighbor
                # We found a corner!
                if (v & GOAL) == 0
                    game.board[□] |= FLAG_ILLEGAL_BOX_TILE # illegal corner
                    if (v & BOX) > 0
                        deadlock = true # box on illegal square
                    end
                end

                types = zero(UInt8)
                if wall_up && wall_left
                    types |= CORNER_TOP_LEFT
                end
                if wall_up && wall_right
                    types |= CORNER_TOP_RIGHT
                end
                if wall_down && wall_left
                    types |= CORNER_BOTTOM_LEFT
                end
                if wall_down && wall_right
                    types |= CORNER_BOTTOM_RIGHT
                end

                ∟ = Corner(□, types, zeros(MVector{N_DIRS, Int32}))
                push!(corners, ∟)
            end
        end
    end

    return (corners, deadlock)
end

"""
Calculate edge-lengths for each corner (edges between corners).
Mark any illegal box tiles on those edges.
"""
function calculate_corner_edge_lengths!(corners::Vector{Corner}, game::Game)

    deadlock = false
    for i in 1:length(corners)-1
        ∟₁ = corners[i]
        col1, row1 = tile_index_to_col_row(game, ∟₁.□)
        for j in i+1 : length(corners) # for each pair of corners
            ∟₂ = corners[j]
            col2, row2 = tile_index_to_col_row(game, ∟₂.□)
            if (col1 == col2) || (row1 == row2) # corners are on same column or on same row
                dx = col2 - col1
                dy = row2 - row1

                # number of squares between the endpoints i and j
                k = max(abs(dx), abs(dy)) - 1

                dx = sign(dx)
                dy = sign(dy)
                dir = dxdy_to_direction(dx, dy)
                a = ∟₁.□

                c, d, c1, d1, box_count, goal_count = 0, 0, 0, 0, 0, 0

                # For each square from corner i to corner ', including the endpoints
                for e in 0:k+1
                    if (game.board[a] & WALL) == 0
                        box_count += (game.board[a] & BOX) > 0
                        goal_count += (game.board[a] & GOAL) > 0

                        a1 = a + game.step_left[dir]
                        a2 = a + game.step_right[dir]
                        if (game.board[a1] & FLOOR) == 0
                            c += 1 # the square on adjacent side 1 is blocked
                        elseif (game.board[a2] & FLOOR) == 0
                            c1 += 1 # opposite side is blocked, i.e., side 2
                        end
                        if (game.board[a2] & FLOOR) == 0
                            d += 1 # the square on adjacent side 2 is blocked
                        elseif (game.board[a1] & FLOOR) == 0
                            d1 += 1 # opposite side is blocked, i.e., side 1
                        end

                        a += game.step_fore[dir]
                    else
                        # non-floor square: the direct line between endpoints i and j is broken
                        break
                    end
                end

                if (c+c1 == k+2) || (d+d1==k+2)
                    # The squares between the endpoints are free, and one of the adjacent sides are blocked,
                    # or each square is blocked by a wall on either side.
                    # Thus, a box can never get away from the line, i.e., it is a closed edge.

                    # Save the line for later calculation of adjacent-line patterns
                    ∟₁.edge_lengths[dir] = k+2

                    if box_count > goal_count
                        deadlock = true # more boxes than goals
                    end

                    if goal_count == 0 # no goals between or at the endpoints
                        for e in 1:k # flag squares as simple illegal squares
                            game.board[∟₁.□ + e*game.step_fore[dir]] |= FLAG_ILLEGAL_BOX_TILE
                        end
                    end

                    if (game.board[∟₂.□] & GOAL) > 0
                        ∟₂.edge_lengths[OPPOSITE_DIRECTION[dir]] = ∟₁.edge_lengths[dir]
                    end
                else
                    # The line between the corners is not a closed edge
                    if a == ∟₂.□ + game.step_fore[dir]
                        # There is an open line between the corners; it is not a closed edge but an open edge from which boxes can escape.
                        ∟₁.edge_lengths[dir] = -(k+2)
                    end
                end
            end
        end
    end

    return deadlock
end

function print_corners(game::Game, corners::Vector{Corner})
    println("\nCorners:")
    chars = [fill(' ', game.board_width+2) for i in 1:game.board_height+2]
    for row in 0:game.board_height+1
        for col in 0:game.board_width+1
            □ = col_row_to_tile_index(game, col, row)
            if (game.board[□] & WALL) > 0
                chars[row+1][col+1] = '#'
            end

            corner_index = findfirst(corner->□==corner.□, corners)
            if !isnothing(corner_index)
                ∟ = corners[corner_index]
                chars[row+1][col+1] = min('a' - 1 + ∟.types, 'z')

                for dir in DIRECTIONS
                    for δ in 1:(∟.edge_lengths[dir]-2)
                        □ = ∟.□ + δ*game.step_fore[dir]
                        col2, row2 = tile_index_to_col_row(game, □)
                        if dir == DIR_UP || dir == DIR_DOWN
                            chars[row2+1][col2+1] = '|'
                        else
                            chars[row2+1][col2+1] = '-'
                        end
                    end
                end
            end
        end
    end

    print_char_matrix(chars)
end