"""
A game defines the static information for a Sokoban puzzle.
It also contains some additional precomputed information.
"""
struct Game
    board_height::Int32 # Number of tiles high that the potentially-player-reachable space is
    board_width::Int32 # Number of tiles across that the potentially-player-reachable space is
                 # Is of length (Width+2)*(Height*2); the extra 2 is for a wall-filled border
    board::Board # The original board used for solving, which has some precomputation / modification

    start_player_tile::TileIndex
    start_board::Board # The original board. Used for display.

    goals::Vector{TileIndex} # The goal squares do not change
    starting_boxes::Vector{TileIndex} # The box starting positions

    step_fore::SVector{N_DIRS, TileIndex} # □ + step_fore[dir] yields the square one step in the given direction from □
    step_left::SVector{N_DIRS, TileIndex} # □ + step_left[dir] is the same as □ + step_fore[rot_left(dir)]
    step_right::SVector{N_DIRS, TileIndex} # □ + step_right[dir] is the same as □ + step_fore[rot_right(dir)]

    hash_components::Vector{UInt64} # Zobrist hash components for box positions
end

tile_indices(game::Game) = TileIndex(1):TileIndex(length(game.board))
box_indices(game::Game) = BoxIndex(1):BoxIndex(length(game.goals))
goal_indices(game::Game) = box_indices(game) # it's the same

"""
Set the goal number bits.
We use the top bits of the goal tiles to their index in `goals` so that
we have a convenient mapping back to the goal number.
This method assumes that the goal number bits have not previously been set.
Typically called after the board has been processed to remove unreachable tiles.
"""
function set_i_goals!(game::Game)
    for (i, □) in enumerate(game.goals)
        game.board[□] += TileValue(i << GOAL_BIT_SHIFT_COUNT)
    end
end

"""
The dynamic state for a Sokoban game, which includes both the lean state information
and the board information for multiple forms of fast analysis.
IE `boxes` lets one quickly identify box locations,
whereas `board` lets one quickly verify if a given location has a box (or player), etc.
This struct is mutable.
"""
mutable struct State
    player::TileIndex
    boxes::Vector{TileIndex}
    board::Board
    zhash::UInt64
end

function State(n_boxes::Int, board_size::Int)
    return State(
        zero(TileIndex),
        Array{TileIndex}(undef, length(n_boxes)),
        Array{TileValue}(undef, length(board_size)),
        zero(UInt64)
    )
end

function State(game::Game, board::Board)
    player = zero(TileIndex)
    boxes = TileIndex[]
    for (□, v) in enumerate(board)
        if (v & PLAYER) > 0
            player = □
        end
        if (v & BOX) > 0
            push!(boxes, □)
        end
    end
    zhash = calculate_zobrist_hash(boxes, game.hash_components)
    return State(player, boxes, copy(board), zhash)
end

"""
Move the player from one location to another.
Player can be removed using □_player=0
"""
function move_player!(s::State, □_player::TileIndex)
    if s.player != 0
        # remove the player from the board
        s.board[s.player] &= ~PLAYER
    end

    # update the player position
    s.player = □_player

    if s.player != 0
        # add the player to the board
        s.board[s.player] |= PLAYER
    end
end

"""
Unset the box flag in the board entry of the given state.
This is useful sometimes when calculating reachability for ideal cases.
"""
function remove_all_boxes_from_board!(s::State)
    for □ in s.boxes
        s.board[□] &= ~BOX
    end
    return s
end
function remove_all_boxes_from_board!(game::Game, s::State)
    for □ in s.boxes
        game.board[□] &= ~BOX
    end
    return game
end

"""
Set the box flag in the board entry of the given state for each box.
"""
function place_all_boxes_on_board!(s::State)
    for □ in s.boxes
        s.board[□] |= BOX
    end
    return s
end
function place_all_boxes_on_board!(game::Game, s::State)
    for □ in s.boxes
        game.board[□] |= BOX
    end
    return game
end


"""
A push is defined as moving one box one step in one direction.
"""
struct Push
    i_box::BoxIndex
    dir::Direction
end

"""
Apply a push to a state, producing a successor state.
We assume the push is valid, meaning that the box is reachable by the player.
The player is placed in the box's location.
NOTE: We do not normalize the resulting player position
"""
function move!(s::State, game::Game, a::Push)
    □ = s.boxes[a.i_box] # where box starts
    ▩ = □ + game.step_fore[a.dir] # where box ends up

    s.board[s.player] &= ~PLAYER # Clear the player
    s.board[□] &= ~BOX # Clear the box
    s.zhash ⊻= game.hash_components[□] # Undo the box's hash component

    s.player = □ # Player ends up where the box is.
    s.boxes[a.i_box] = ▩

    s.board[s.player] |= PLAYER # Add the player
    s.board[▩] |= BOX # Add the box
    s.zhash ⊻= game.hash_components[▩] # Add the box's hash component

    return s
end

"""
Reverse a push on a state, producing the predecessor state.
We assume that the push is valid.
The player is placed in the location prior to push.
NOTE: We do not normalize the resulting player position.
"""
function unmove!(s::State, game::Game, a::Push)
    □ = s.boxes[a.i_box] # where box starts
    ▩ = □ - game.step_fore[a.dir] # where box ends up

    s.board[s.player] &= ~PLAYER # Clear the player
    s.board[□] &= ~BOX # Clear the box
    s.zhash ⊻= game.hash_components[□] # Undo the box's hash component

    s.player = ▩ - game.step_fore[a.dir] # Player ends up behind the box.
    s.boxes[a.i_box] = ▩

    s.board[s.player] |= PLAYER # Add the player
    s.board[▩] |= BOX # Add the box
    s.zhash ⊻= game.hash_components[▩] # Add the box's hash component

    return s
end

"""
Move the player one step in the given direction.
If there is a box, push the box.
This assumes that the move is valid.
This method does not assume that the player is in a normalized location.
"""
function move!(s::State, game::Game, dir::Direction)
    □ = s.player # where the player starts
    ▩ = □ + game.step_fore[dir] # where the player ends up

    s.board[□] &= ~PLAYER # Clear the player
    s.player = ▩ # Where the player ends up.
    s.board[s.player] |= PLAYER # Add the player

    if (s.board[▩] & BOX) > 0
        # Assume this is a valid push
        # We do not check that the next tile over exists, is empty, etc.
        ◰ = ▩ + game.step_fore[dir] # where box ends up

        # This should always succeed
        i_box = findfirst(isequal(▩), s.boxes)

        s.board[▩] &= ~BOX # Clear the box
        s.zhash ⊻= game.hash_components[▩] # Undo the box's hash component

        s.boxes[i_box] = ◰
        s.board[◰] |= BOX # Add the box
        s.zhash ⊻= game.hash_components[◰] # Add the box's hash component
    end

    return s
end


"""
Returns true if our game has been solved.
IE every box is on a goal.
"""
function is_solved(game::Game, s::State)
    for □ in s.boxes
        if (game.board[□] & GOAL) == 0
            return false
        end
    end
    return true
end

"""
col=1, row=1 will give us the top-left non-border square.
col=2, row=1 will give us the square just right of that, etc.
"""
function col_row_to_tile_index(game::Game, col::Integer, row::Integer)::TileIndex
    return col_row_to_tile_index(game.board_width, col, row)
end

function tile_index_to_col_row(game::Game, □::Integer)::Tuple{Int,Int}
    colrow = CartesianIndices((game.board_width+2, game.board_height+2))[□]
    return (colrow[1]-1, colrow[2]-1)
end

function is_tile_within_border(game::Game, □::Integer)
    col, row = tile_index_to_col_row(game, □)
    return col ≥ 1 && col ≤ game.board_width && row ≥ 1 && row ≤ game.board_height
end

"""
Get the Manhattan distance between two tiles
"""
function get_manhattan_dist(game::Game, □::TileIndex, ▩::TileIndex)
    □_col, □_row = tile_index_to_col_row(game, □)
    ▩_col, ▩_row = tile_index_to_col_row(game, ▩)
    return abs(□_col - ▩_col) + abs(□_row - ▩_row)
end

"""
Get the minimum distance to a tile based on a push or pull distances matrix.
Returns typemax(Int32) if the tile cannot be reached.
"""
function get_push_or_pull_dist(
    □::TileIndex,
    distances_by_direction::Matrix{Int32}, # Of size board_size × N_DIRS
    )::Int32
    res = typemax(Int32)
    for dir in DIRECTIONS
        d = distances_by_direction[□, dir]
        if (d < res) && (d != typemin(Int32))
            res = d
        end
    end
    return res
end

function board_to_text(game::Game, board::Board)::String
    result = ""
    for row in 0:game.board_height+1
        for col in 0:game.board_width+1
            v = board[col_row_to_tile_index(game, col, row)]
            result *= square_to_char(v)
        end
        result *= "\n"
    end
    return result
end

function load_game_from_string(str::String)

    # Figure out the board height and width,
    # and verify that the board width does not change
    running_board_width = zero(Int32)
    board_width = zero(Int32)
    board_height = one(Int32)
    for c in str
        if c == '\n'
            if board_width == 0
                board_width = running_board_width
            else
                @assert board_width == running_board_width
            end
            running_board_width = zero(Int32)
            board_height += one(Int32)
        else
            running_board_width += one(Int32)
        end
    end

    # Get the inner dimensions
    board_width -= Int32(2)
    board_height -= Int32(2)

    # Allocate the boards
    board_size = (board_width+2)*(board_height+2)
    start_board = Array{TileValue}(undef, board_size)

    # Put a border around the actual board we use
    board = Array{TileValue}(undef, board_size)
    let
        # top row
        for i in 1:board_width+2
            board[i] = WALL
        end

        # bottom row
        for i in board_size : -1 : board_size - board_width - 1
            board[i] = WALL
        end

        # left and right columns
        row_offset = 1
        for i in 1:board_height
            row_offset += board_width+2
            board[row_offset] = WALL
            board[row_offset+board_width+1] = WALL
        end
    end

    # Now populate the board and compute our starting values
    start_player_tile = zero(TileIndex)
    goals = TileIndex[]
    boxes = TileIndex[]
    let
        row = 0
        col = 0
        for c in str
            v = char_to_square(c)
            if isnothing(v)
                row += 1
                col = 0
                continue
            end

            □ = col_row_to_tile_index(board_width, col, row)
            start_board[□] = v

            # Only set the value if we are within the border
            if 1 ≤ row ≤ board_height && 1 ≤ col ≤ board_width
                board[□] = v
                # Remember starting values
                if (v & PLAYER) > 0
                    start_player_tile = □
                end
                if (v & GOAL) > 0
                    push!(goals, □)
                end
                if (v & BOX) > 0
                    board[□] |= FLAG_BOX_START_TILE
                    push!(boxes, □)
                end
            end

            col += 1
        end
    end

    # Construct our steps
    step_fore = SVector{N_DIRS, TileIndex}(-(board_width+2), -1, board_width+2, 1)
    step_left = SVector{N_DIRS, TileIndex}(step_fore[2], step_fore[3], step_fore[4], step_fore[1])
    step_right = SVector{N_DIRS, TileIndex}(step_fore[4], step_fore[1], step_fore[2], step_fore[3])

    # Construct our hash components
    hash_components = construct_zobrist_hash_components(board_size)

    return Game(
        board_height,
        board_width,
        board,
        start_player_tile,
        start_board,
        goals,
        boxes,
        step_fore,
        step_left,
        step_right,
        hash_components
        )
end

function load_game_from_strings(strings::Vector{String})
    max_len = maximum(length(str) for str in strings)

    big_string = ""
    for (str_index,str) in enumerate(strings)
        big_string *= str
        n_chars_to_add = max_len-length(str)
        for i in 1:n_chars_to_add
            big_string *= " "
        end
        if str_index != length(strings)
            big_string *= "\n"
        end
    end

    return load_game_from_string(big_string)
end
