# In the final phase of the calculation, boxes are allowed to be removed from the board via a box
# starting position even if the distance to the starting position is below the normal distance threshold
const END_GAME_PHASE_PERCENT = 85

const LARGE_BOARD_AREA_THRESHOLD_PERCENT = 25

# Maximum park-box attempts for the packing order calculation; the number should be small enough to guarantee a reasonably small running time upper bound
const MAX_PACKING_ORDER_PARK_BOXES_ATTEMPTS = 500 # 500*1000

@enum PackingOrderType begin
    PACKING_ORDER_PULL_BOXES
    PACKING_ORDER_PULL_BOXES_TO_STARTING_POSITIONS_WITH_PARKING
    PACKING_ORDER_PULL_BOXES_TO_STARTING_POSITIONS_WITHOUT_PARKING
end

"""
A PackingOrder struct represents a series of phases in which we move boxes to goals, such
that once all phases are complete, the game is solved.
Each phase has a "set number". If two goals have the same set number, then boxes are pushed to them at the same time.

We find a packing order via packing order search. This starts with all boxes on their goals, and tries to move boxes
to their starting positions. As such, it will first find set_number=1, then set_number=2. Thus, actually solving
the game requires executing the set numbers / phases in decreasing order.

This struct both captures the final packing order discovered, but also contains some extra memory used
in packing order search.

A parking tile is an intermediate goal. That is, boxes may be temporarily moves to a parking tile before being moved to
a goal, in order to be leave room for other boxes.
"""
mutable struct PackingOrder
    # The tile indices for the goal and parking tiles.
    # The first tiles are the goals, followed by any parking tiles.
    # Preallocated to a length N ≥ |goals|.
    goal_and_parking_tiles::Vector{TileIndex}

    # The number of goal and parking tiles currently set in packing order search.
    # |goals| ≤ goal_and_parking_tile_count ≤ |goal_and_parking_tiles|
    goal_and_parking_tile_count::Int

    # For each goal (or parking tile), its set of reachable box starting positions
    # The set of boxes that it can consider.
    # Preallocated to length N.
    goal_box_sets::Vector{Vector{BoxIndex}}

    # Number of populated goal box sets.
    # i.e., number of game phases when the packing order is used by the search for a solution.
    set_count::Int

    # Circular linked lists for goals/parking tiles sharing the same tile;
    # a tile can at the same time be a real goal square and used as parking space one or more times
    # Preallocated to length N.
    next_goal_at_same_square::Vector{BoxIndex}

    # The goals and parking tiles for each phase of the packing order.
    # e.g., sets A,B, and C with 2, 3, and 2 members look like: [a1,a2,b1,b2,b3,c1,c2] (vector element 0 is unused)
    #                                                            ^     ^        ^     ^
    #                                                            1     3        6     8
    # The corresponding 'first_set_member_index' vector is: [1,3,6,8]
    # Note that 8 is a sentinel value such that first_set_member_index[i] : first_set_member_index[i+1]-1 can be used to
    # get all goals for a given phase.
    # Preallocated to length N+1.
    set_members::Vector{BoxIndex}

    # The index of first goal in `set_members` for each set number.
    # Preallocated to length N.
    first_set_member_index::Vector{Int32}

    # A reverse mapping of set_members.
    # That is, a quick lookup from goal/parking tile to its packing set number.
    # Preallocated to length N.
    goal_set_numbers::Vector{UInt8}

    # For each parked box, the packing set index of the goal the parked box came from.
    # A parked box should preferably stay at its parking tile until it is time to fill the goal the parked box originally came from.
    # Preallocated to length N.
    parked_box_destination_goal_set_numbers::Vector{UInt8}

    # The last (lowest) game phase where a box is parked because a goal square
    # must be filled with a box coming from a relatively small area of the board.
    last_packing_set_with_a_box_parking_in_a_small_area::UInt8

    # Distances from each tile to each goal or parking tile,
    # truncated to 254, where typemax(UInt8) indicates no path to the tile.
    # (of size MAX_BOARD_SIZE×MAX_BOX_NUMBER)
    tile_goal_distances::Matrix{UInt8}

    # calculation time [s]
    calc_time::Float64
end

goal_and_parking_numbers(goal_and_parking_tile_count::Integer) = one(BoxIndex):BoxIndex(goal_and_parking_tile_count)
goal_and_parking_numbers(packing_order::PackingOrder) = one(BoxIndex):BoxIndex(packing_order.goal_and_parking_tile_count)

"""
A goal number refers to a goal if it is 1:|goals|.
Larger values refer to parking tiles.
"""
is_real_goal(i_goal::Integer, game::Game) = i_goal ≤ length(game.goals)
is_parking_tile(i_goal::Integer, game::Game) = i_goal > length(game.goals)

function print_packing_order(game::Game, packing_order::PackingOrder)
    chars = [fill(' ', game.board_width+2) for i in 1:game.board_height+2]
    for row in 0:game.board_height+1
        for col in 0:game.board_width+1
            □ = col_row_to_tile_index(game, col, row)
            v = game.board[□]
            i_goal = get_i_goal(v)
            if ((v & GOAL) == 0) &&
               (i_goal == 0)
                if (v & WALL) > 0
                    chars[row+1][col+1] = '#'
                end
            else
                set_number = packing_order.goal_set_numbers[i_goal]
                if set_number != 0
                    chars[row+1][col+1] = Char('0' + set_number)
                    if packing_order.next_goal_at_same_square[i_goal] != i_goal
                        # Multiple go to this square
                        chars[row+1][col+1] = '*'
                    end
                end
            end
        end
    end

    print_char_matrix(chars)
end


"""
Returns true if the remaining boxes on the board are located at the box starting positions
"""
function all_boxes_at_starting_positions(packing_order::PackingOrder, s::State)
    for i_goal in goal_and_parking_numbers(packing_order)
        □ = packing_order.goal_and_parking_tiles[i_goal]
        if s.board[□] & (BOX + FLAG_BOX_START_TILE) == BOX
            # A goal with a box that is not in its start position
            return false
        end
    end
    return true
end

"""
This method calculates both:
    tile_goal_distances - the distance of each tile to each goal
    goal_box_sets - for each goal, its set of reachable box starting positions
It is basically #7167, but we always use the packing order.

Our first call to this method will be before we have any parking squares.
As such, we will call it with i_goal_lo = 1, i_goal_hi = length(goals).
Later, as we add parking squares, we will add those separately, and call
this with i_goal_lo > length(goals).
"""
function calc_tile_goal_distances_and_goal_box_sets!(
    tile_goal_distances::Matrix{UInt8}, # of size n_tiles × n_goals. Note that the type is UInt8
    goal_and_parking_tiles::Vector{TileIndex}, # (from PackingOrder)
    goal_box_sets::Vector{Vector{UInt8}},
    goal_set_numbers::Vector{UInt8}, # mapping of each goal number (or parking square) to its packing order set number - ie its phase
    parked_box_destination_goal_set_numbers::Vector{UInt8}, # for each parked box, the set number of the goal the parked box came from; a parked box should preferably stay at its parking square until it's time to fill the goal the parked box originally came from
    i_goal_lo::BoxIndex,
    i_goal_hi::BoxIndex,
    destination_offset::Int,
    game::Game,
    s::State,
    freeze_data::ReachableTiles,
    distances_by_direction::Matrix{Int32},
    reach::ReachableTiles,
    stack::Vector{TileIndex},
    queue::Vector{Tuple{TileIndex, TileIndex, Int32}}
    )

    # Get the state ready
    □_player_orig = s.player
    remove_all_boxes_from_board!(s)

    for i_goal in i_goal_lo : i_goal_hi
        # Place boxes according to the packing order.
        # IE all boxes should be in the positions we expect them to be when we get to this goal.
        # Only consider real goal tiles, not parking tiles, as boxes on parking tiles can move -
        # that is the point of parking tiles.
        target_set_number = goal_set_numbers[i_goal]
        for g in box_indices(game)
            # NOTE: This is this > because larger phases happen first.
            if (goal_set_numbers[g] > target_set_number) &&
               (parked_box_destination_goal_set_numbers[g] ≤ target_set_number)
                s.board[game.goals[g]] |= BOX
            end
        end

        # Remove non-freezing boxes so that we find all squares that can
        # be reached by pulling a box around the board, starting from `i_goal`.
        for g in box_indices(game)
            if ((s.board[game.goals[g]] & BOX) != 0) &&
               (!is_a_freezing_move(
                    one(TileIndex), # src
                    game.goals[g], # dst
                    game,
                    s,
                    freeze_data,
                    true)) # a box freezing on a goal square does count as a freezing move
                s.board[game.goals[g]] &= ~BOX
            end
        end

        calc_box_pull_or_push_distances!(
            distances_by_direction, game, s, packing_order.goal_and_parking_tiles[i_goal],
            reach, stack, queue,
            false, false, false)

        # TODO: Should be iterating over boal goal *and parking tile count*
        for g in box_indices(game)
            s.board[game.goals[g]] &= ~BOX
        end

        for □ in tile_indices(game)
            # Initialize to typemax, meaning no path from the square to the goal
            tile_goal_distances[□, destination_offset + i_goal] = typemax(eltype(tile_goal_distances))
            if is_a_legal_and_box_reachable_tile(s.board[□])
                for dir in DIRECTIONS
                    if !is_type_extrema(distances_by_direction[□, dir])
                        tile_goal_distances[□, destination_offset + i_goal] =
                            min(
                                tile_goal_distances[□, destination_offset + i_goal],
                                distances_by_direction[□,dir],
                                typemax(eltype(distances))-1
                            )
                    end
                end
            end
        end

        # For each goal, calculate the box starting positions that are reachable
        # by pulling a box from the goal square to the box starting position,
        # as if playing the game in reverse mode.
        empty!(goal_box_sets[destination_offset + i_goal])
        for i_box in box_indices(game)
            square_goal_dist = tile_goal_distances[game.starting_boxes[i_box], destination_offset+i_goal]
            if square_goal_dist < typemax(eltype(tile_goal_distances))
                # A box can be pulled from the goal `i_goal` to the box starting position with index 'i_box'
                push!(goal_box_sets[destination_offset + i_goal], i_box)
            end
        end
    end

    place_all_boxes_on_board!(game, s)
    s.player = □_player_orig

    # This method has updated both tile_goal_distances and goal_box_sets
    nothing
end

struct CanPullBoxToStartingPositionResult
    success::Bool                           # If true, we can pull the box to a starting position
    box_start_number::BoxIndex             # Index of the box in original game state list of boxes, which is the starting position we pull to
    □_player_new::TileIndex                 # Tile index that the player ends up at
    n_pulls::Int                            # Number of moves the player took to complete the pull
    n_reachable_box_starting_positions::Int # Number of reachable box starting positions, same as resulting value of length(reachable_box_starting_positions)
    n_reachable_tiles::Int                  # Number of reachable tiles
end

# 13210
"""
Determine whether we can pull a box from a goal square to a starting position.
This computes box pull distances using the given preallocated distances.
This method also clears and computes reachable_box_starting_positions.
"""
function can_pull_box_to_starting_position(
    i_goal::BoxIndex,
    game::Game,
    s::State,
    reach::ReachableTiles,
    stack::Vector{TileIndex},
    queue::Vector{Tuple{TileIndex, TileIndex, Int32}},
    distances_by_direction::Matrix{Int32}, # Of size board_size × N_DIRS
    reachable_box_starting_positions::Vector{BoxIndex},
    goal_and_parking_tiles::Vector{TileIndex}, # (from PackingOrder)
    goal_and_parking_tile_count::Integer, # (from PackingOrder) total number of goals and parking tiles
    goal_box_sets::Vector{Vector{UInt8}}, # (from PackingOrder) for each goal (or parking tile), the set of reachable box starting positions from the goal - should technically be sets
    goal_set_numbers::Vector{UInt8}, # (from PackingOrder) mapping of each goal number (or parking tile) to its packing order set number - ie its phase
    set_members::Vector{UInt8}, #  (from PackingOrder) goal sets, e.g., sets A,B, and C with 2, 3, and 2 members look like: [0,a1,a2,b1,b2,b3,c1,c2] (vector element 0 is unused); the corresponding 'FirstSetMemberIndex' vector is: [0,1,3,6,8]; '8' is a sentinel
    min_dist_between_box_and_removal_tile::Int = 3 # Manhattan distance threshold; if a box is removed from the board by using a too close starting square, then removal of the remaining boxes may look easier than it really is
    )::CanPullBoxToStartingPositionResult

    success = false
    box_start_number = zero(BoxIndex)
    □_player_new = zero(TileIndex)
    n_pulls = 0
    n_reachable_tiles = 0

    □_goal = goal_and_parking_tiles[i_goal]

    calc_box_pull_or_push_distances!(
        distances_by_direction,
        game,
        s,
        □_goal,
        reach,
        stack,
        queue,
        false, # pull the box
        true, # the player's current position matters when initial legal pulls are generated
        false # we are not continuing the calculation
    )
    n_reachable_tiles = count_reachable_tiles(distances_by_direction)

    □_player_orig = s.player

    # Remove the box, if any, from the board
    s.board[□_goal] &= ~BOX

    empty!(reachable_box_starting_positions)

    # We now conduct a greedy search for a target box starting position.
    # This search tries to leave chances open for the remaining goals by maximizing the number of reachable box starting positions.
    max_n_reachable_box_starting_positions = -1
    for i_box in box_indices(game)
        # NOTE: We always start the loop with the box in its correct position and the player on their original tile

        □_box_start = game.starting_boxes[i_box]
        box_can_be_considered = true

        if ((s.board[□_box_start] & FLAG_VISITED_TILE) == 0) && # The box starting position has not been used for removing another box
           ((get_manhattan_dist(game, □_box_start, □_goal) ≥ min_dist_between_box_and_removal_tile) # The box and the removal square cannot be too close; otherwise, if it's a tightly packed goal room then using this square for removal could make the removal of the remaining boxes easier than it really is
             ||
            ( # Allow the box to be removed from the board using its current square if it's an original goal square and not a parking tile; it may be a frozen box
             (□_goal == □_box_start) && # the box is already at a starting square
             (i_goal ≤ length(game.goals)) # it is a goal square and not a parking space
            )
            #  || TODO
            # (remaining_boxes_count <= end_game_phase_box_count_limit) # most of the boxes, including parked boxes, have been removed from the board
           )

            min_n_reachable_box_starting_positions = typemax(Int)
            if □_box_start != □_goal
                # Box is not already at the starting position
                for dir in DIRECTIONS
                    if !is_type_extrema(distances_by_direction[□_box_start, dir])
                        # A box on the goal can be pulled to this box starting position

                        if box_can_be_considered
                            box_can_be_considered = false
                            push!(reachable_box_starting_positions, i_box)
                        end

                        # Check whether the pusher, after pulling the box to this box start position,
                        # can get back to where they came from. This heuristic is used to accept the box
                        # starting square as a target for the pull. We auto-accept if it is the last box.
                        pull_is_accepted = true
                        □_player = □_box_start + game.step_fore[dir] # where the player ends up after the pull
                        if length(set_members) < goal_and_parking_tile_count - 1
                            # This is not the last box
                            s.board[□_box_start] |= BOX # Move the box here
                            calc_reachable_tiles!(reach, game, s.board, □_player, stack)
                            s.board[□_box_start] &= ~BOX # Remove the box
                            pull_is_accepted = is_reachable(□_player_orig, reach)
                        end

                        if pull_is_accepted
                            for i_goal in 1:length(goal_box_sets)
                                n_reachable_box_starting_positions = length(goal_box_sets[i_goal])
                                if (n_reachable_box_starting_positions < min_n_reachable_box_starting_positions) &&
                                   (goal_set_numbers[i_goal] == 0) && # the goal or parking tile hasn't been assigned a packing order set number yet
                                   (i_box ∈ goal_box_sets[i_goal]) # the box starting position is pull-reachable from this goal (TODO: goal_box_sets is not calculated!)
                                    min_n_reachable_box_starting_positions = n_reachable_box_starting_positions
                                end
                            end
                            if min_n_reachable_box_starting_positions > max_n_reachable_box_starting_positions
                                # i_box is a new/first best box starting position candidate.
                                # the best candidate is the one with MAX(MIN(still open goal-box edges))}
                                success = true
                                box_start_number = i_box
                                □_player_new = □_player
                                max_n_reachable_box_starting_positions = min_n_reachable_box_starting_positions # remember the best score
                                n_pulls = distances_by_direction[□_box_start, dir]
                            end
                            break # exit the for loop
                        end
                    end
                end
            else
                # The box is already at a box starting position
                for i_goal in 1:length(goal_box_sets)
                    n_reachable_box_starting_positions = length(goal_box_sets[i_goal])
                    if (n_reachable_box_starting_positions < min_n_reachable_box_starting_positions) &&
                       (goal_set_numbers[i_goal] == 0) && # the goal or parking tile hasn't been assigned a packing order set number
                       (i_box ∈ goal_box_sets[i_goal]) # the box starting position is pull-reachable from this goal (TODO: goal_box_sets is not calculated!)
                        min_n_reachable_box_starting_positions = length(goal_box_sets[i_goal])
                    end
                end
                if min_n_reachable_box_starting_positions > max_n_reachable_box_starting_positions
                    # i_box is a new/first best box starting position candidate; the best candidate is the one with MAX(MIN(still open goal-box edges))
                    success = true
                    box_start_number = i_box
                    □_player_new = s.player # save the player position
                    max_n_reachable_box_starting_positions = min_n_reachable_box_starting_positions # remember the best score
                    n_pulls = 0
                end

                box_can_be_considered = false
                push!(reachable_box_starting_positions, i_box)
            end
        else
            # This box starting position is not a candidate for removing the box at the given goal
            for dir in DIRECTIONS
                if !is_type_extrema(distances_by_direction[□_box_start, dir])
                    # The box at the given goal can be pulled to this box starting position
                    if box_can_be_considered
                        box_can_be_considered = false
                        break
                    end
                end
            end
        end

        # Reset the board with the box on the square and the player on their original tile
        s.board[□_goal] |= BOX
        move_player!(s, □_player_orig)
    end

    return CanPullBoxToStartingPositionResult(success, box_start_number, □_player_new, n_pulls, length(reachable_box_starting_positions), n_reachable_tiles)
end

struct ParkingSpace
    distance::Int
    □::TileIndex
    □_player::TileIndex # after parking
end

function is_valid(parking_space::ParkingSpace)
    return parking_space.□ > 0
end

# 12981
"""
Finds at most two parking spaces to stash boxes at.
Always returns a tuple of two parking spaces. Their distances are less than zero
if they are invalid.

The first parking square is the one furthest from the source tile index,
as measured by the push distance on an empty board. These square-to-goal
distances are recomputed and available in `packing_order.tile_goal_distances`.

The secondary parking square is the one furthest from the goal,
as measured by the Manhattan distance. The Manhattan distance between
the primary and second parking square is used as a tie breaker.
"""
function find_greatest_distance_parking_candidates(
    □_box::TileIndex,
    game::Game,
    s::State,
    distances_by_direction::Matrix{Int32}, # Of size board_size × N_DIRS
    reach::ReachableTiles,
    stack::Vector{TileIndex},
    queue::Vector{Tuple{TileIndex, TileIndex, Int32}},
    tile_goal_distances::Matrix{UInt8}, # of size n_tiles × n_goals (typically from packing order)
    )::Tuple{ParkingSpace, ParkingSpace}

    calc_box_pull_or_push_distances!(
            distances_by_direction, game, s, □_box,
            reach, stack, queue,
            false, # pull the box
            true, # use the player position
            false # we are not continuing the calculation
            )

    # Save the player location and remove the box
    □_player_orig = s.player
    s.board[□_box] &= ~BOX

    # We just calculated the pull distances from the goal wrt the current game state.
    # Now we find the greatest push parking distances from the goal.
    greatest_dist_push = Int32(-1)
    greatest_dist_manhattan = Int32(-1)
    box_push_back_to_goal_distances = similar(s.board, Int32)
    for □ in tile_indices(game)
        distance = get_push_or_pull_dist(□, distances_by_direction)
        if distance != typemax(Int32)
            # The box can be pulled from the goal to □
            if i_goal ≤ length(game.goals)
                # It is a normal goal and not a parking space
                if tile_goal_distances[□, i_goal] < typemax(eltype(tile_goal_distances))
                    # The distance to the goal does not exceed the (byte-sized) maximum distance
                    distance = tile_goal_distances[□, i_goal]
                else
                    distance = typemax(Int32)
                end
            else
                # It is a parking space, i.e., a pseudo-goal.
                # We don't have tile → goal distances precomputed for parking spaces.
                # Only allow the parked box to move a few steps from its current location.
                if distance > 2
                    distance = typemax(Int32)
                end
            end
        end
        box_push_back_to_goal_distances[□] = distance

        # Now update distances_by_direction with the newly computed distance
        if !is_type_extrema(distance)
            for dir in DIRECTIONS

                if !is_type_extrema(distances_by_direction[□, dir])
                    distances_by_direction[□, dir] = distance
                end
            end

            greatest_dist_push = max(greatest_dist_push, distance)

            distance = get_manhattan_dist(game, □, □_box)
            greatest_dist_manhattan = max(greatest_dist_manhattan, distance)
        end
    end

    # Find the parking spaces that maximize the distance metrics.
    parking_space_greatest_push = ParkingSpace(greatest_dist_push, zero(TileIndex), zero(TileIndex))
    parking_space_greatest_manhattan = ParkingSpace(greatest_dist_manhattan, zero(TileIndex), zero(TileIndex))

    # Search until we find two parking spaces, or until we've exhausted our search.
    n_parking_spaces_found = 0
    while (n_parking_spaces_found < 2) && (greatest_dist_push > 0) && (greatest_dist_manhattan > 0)
        found_a_tile_this_iteration = false
        □ = zero(TileIndex)
        while □ < length(s.board)
            □ += one(TileIndex)

            # Only choose box-reachable tiles that are not illegal, not gate tiles, and not tunnel tiles.
            d_manhattan = get_manhattan_dist(game, □, □_box)
            if ((s.board[□] & (FLAG_BOX_REACHABLE_TILE +
                               FLAG_ILLEGAL_BOX_TILE +
                               FLAG_GATE_TILE +
                               DIRECTION_TO_TUNNEL_FLAGS)) == FLAG_BOX_REACHABLE_TILE) &&
               (
                 (
                    (
                        # We first check for the greatest-push-distance parking tile
                        (n_parking_spaces_found == 0) &&
                        (box_push_back_to_goal_distances[□] == greatest_dist_push)
                    )
                    ||
                    (
                        # Next we check for the greatest-manhattan-distance parking tile
                        (n_parking_spaces_found == 1) &&
                        (d_manhattan == greatest_dist_manhattan) &&
                        (□ != parking_space_greatest_push.□) # manhattan tile is different than push tile
                    )
                    ||
                    (
                        # If we are still searching, we try to imporve our greatest-manhattan-distance
                        # tile by finding one that is furthest from our greatest-push-distance parking tile
                        (n_parking_spaces_found == 2) &&
                        (d_manhattan == greatest_dist_manhattan) &&
                        (get_manhattan_dist(game, □, parking_space_greatest_push.□) >
                         get_manhattan_dist(game, parking_space_greatest_manhattan.□, parking_space_greatest_push.□))
                    )
                  )
               )
                # The □ is a parking square candidate.
                # Ensure that the player can get back to where they came from.
                # This is a heuristic for better parking tiles.
                # Here `dir` is the side of the tile that the player ends up after pulling to the tile.
                for dir in DIRECTIONS
                    if □ ≤ length(s.board) && # This first condition lets us exit the for loop as soon as we find the primary parking tile
                       !is_type_extrema(distances_by_direction[□, dir])
                        # Put the box on the tile
                        s.board[□] |= BOX

                        □_player_after = □ + game.step_fore[dir]
                        move_player!(s, □_player_after)
                        calc_reachable_tiles!(reach, game, s.board, □_player_after, stack)

                        # Remove the box from the board again
                        s.board[□] &= ~BOX

                        if is_reachable(□_box, reach)
                            found_a_tile_this_iteration = true
                            if n_parking_spaces_found < 2
                                n_parking_spaces_found += 1
                            end
                            if n_parking_spaces_found == 1
                                parking_space_greatest_push = ParkingSpace(greatest_dist_push, □, □_player_after)
                                # We found the primary parking tile - exit the for loops.
                                □ = TileIndex(length(s.board) + 1)
                            else
                                parking_space_greatest_manhattan = ParkingSpace(greatest_dist_manhattan, □, □_player_after)
                            end
                        end
                    end
                end
            end
        end

        if !found_a_tile_this_iteration
            # No parking tile was found with the currently tested distance.
            # Prepare for the next round by decrementing our max distance.
            if n_parking_spaces_found == 0
                greatest_dist_push -= 1
                parking_space_greatest_push = ParkingSpace(greatest_dist_push, zero(TileIndex), zero(TileIndex))
            else
                greatest_dist_manhattan -= 1
                parking_space_greatest_manhattan = ParkingSpace(greatest_dist_manhattan, zero(TileIndex), zero(TileIndex))
            end
        end
    end

    # Put the box back on the board and restore the player position
    s.board[□_box] |= BOX
    move_player!(s, □_player_orig)

    return (parking_space_greatest_push, parking_space_greatest_manhattan)
end

mutable struct PackingOrderSearchData
    count::Int # highest packing order set number

    # Keep track of the best partial packing order
    best_goal_set_count::Int
    best_goal_set_remaining_boxes_count::Int
    best_goal_set_numbers::Vector{BoxIndex}
    goal_group_numbers::Vector{BoxIndex} # TODO: Verify type (for each goal, the group of connected goals to which it belongs)

    set_members_count::Int # How many phases of packing order search have occurred?
    remaining_boxes_count::Int # The number of boxes left to get to goals
    remaining_real_goals_count::Int
    goal_parking_tiles::Vector{TileIndex}
    parked_twice_goal_set::Vector{BoxIndex} # goal number
    seen_box_start_positions::Vector{TileIndex} # technically a set
    large_board_area_threshold::Int
    park_boxes_down_count::Int # Starts at MAX_PACKING_ORDER_PARK_BOXES_ATTEMPTS and counts down
end

# 13355
function add_goal_removal_candidate!(
    goal_removal_candidates::Vector{BoxIndex},
    goal_removal_candidates_box_start_numbers::Vector{BoxIndex},
    used_box_start_numbers::Vector{BoxIndex},
    i_goal::BoxIndex,
    packing_order::PackingOrder,
    data::PackingOrderSearchData,
    packing_order_type::PackingOrderType,
    box_start_number::BoxIndex, # Index of the box in original game state list of boxes, which is the starting position we pull to
    game::Game,
    s::State,
    set_number::Int, # Number of the set to which we are adding the goal
    )
    # Add the goal to the set and save the start number
    push!(goal_removal_candidates, i_goal)
    push!(goal_removal_candidates_box_start_numbers, box_start_number)
    push!(used_box_start_numbers, box_start_number)

    # Tentatively assign a packing order set number to the goal
    packing_order.goal_set_numbers[i_goal] = set_number

    if packing_order_type == PACKING_ORDER_PULL_BOXES_TO_STARTING_POSITIONS_WITHOUT_PARKING
        # Mark the box starting position as 'used'
        s.board[game.starting_boxes[box_start_number]] |= FLAG_VISITED_TILE

        # Update the number of remaining unused box starting positions for all boxes
        # on the board, i.e., boxes at goals and parking tiles.
        for i in goal_and_parking_numbers(packing_order)
            goal_box_set = packing_order.goal_box_sets[i]
            if box_start_number ∈ goal_box_set
                delete!(goal_box_set, box_start_number)
            end
       end
    end

    if is_real_goal(i_goal, game)
        # We removed a real goal
        data.remaining_real_goals_count -= 1
    end
end

# 13374
function clear_removal_candidate_set!(
    goal_removal_candidates::Vector{BoxIndex},
    goal_removal_candidates_box_start_numbers::Vector{BoxIndex},
    packing_order::PackingOrder,
    data::PackingOrderSearchData,
    packing_order_type::PackingOrderType,
    game::Game,
    s::State,
)
    while !isempty(goal_removal_candidates)
        i_goal = goal_removal_candidates[end]
        staring_i_box = goal_removal_candidates_box_start_numbers[length(goal_removal_candidates)]
        pop!(goal_removal_candidates)
        packing_order.goal_set_numbers[i_goal] = zero(UInt8) # clear the tentatively assigned packing order set number for the goal

        # TODO: Technically ≥
        if packing_order_type == PACKING_ORDER_PULL_BOXES_TO_STARTING_POSITIONS_WITHOUT_PARKING
            # Clear the 'used' mark for the box starting position
            s.board[s.boxes[staring_i_box]] &= ~FLAG_VISITED_TILE

            # Update the number of remaining unused box starting positions for
            # all boxes on the board, i.e., boxes at goals and parking tiles.
            for i_i_goal in goal_and_parking_numbers(packing_order)
                if staring_i_box ∈ goal_box_sets[i_i_goal]
                    data.count += 1
                end
            end
        end
        if i_goal ≤ length(game.goals)
            data.remaining_real_goals_count += 1
        end
    end
end

# 13136
"""
Precondition: the player's reachable squares have already been calculated
"""
function find_goal_candidates_for_parking(
    game::Game,
    s::State,
    distances_by_direction::Matrix{Int32}, # Of size board_size × N_DIRS
    reach::ReachableTiles,
    stack::Vector{TileIndex},
    queue::Vector{Tuple{TileIndex, TileIndex, Int32}},
    packing_order::PackingOrder,
    data::PackingOrderSearchData,
    tested_goals_set::Vector{BoxIndex}
    )::Tuple{BoxIndex, Tuple{ParkingSpace, ParkingSpace}} # goal number and parking spaces

    best_i_goal = zero(BoxIndex)
    best_parking_spaces = (
        ParkingSpace(zero(Int32), zero(TileIndex), zero(TileIndex)),
        ParkingSpace(zero(Int32), zero(TileIndex), zero(TileIndex))
    )

    # Search over all goals and parking tiles
    for i_goal in goal_and_parking_numbers(packing_order)
        □ = packing_order.goal_and_parking_tiles[i_goal]
        if (packing_order.goal_set_numbers[i_goal] == 0) && # the goal has not been assigned a packing order set number
           (i_goal ∉ tested_goals_set) &&
           (is_real_goal(i_goal, game)) # This is a real goal, not a parking space
           #  or
           #  ((ParkedTwiceGoalSet__.Count<MAX_PARKED_TWICE_GOAL_COUNT) {'True': the maximum number of parked-twice boxes hasn't been exceeded}
           #   and
           #   (not (GoalNo in ParkedTwiceGoalSet__.BoxSet))
           #  )
           # )
           (data.park_boxes_down_count > 0) # the maximum number of parking attempts has not been exceeded

            @assert (s.board[□] & BOX) > 0 # sanity check - box should always be there
            is_a_corner_goal_tile = is_a_corner(game, □)

            for dir in DIRECTIONS
                # The neighbor square is an empty floor tile, and it is legal to put a box there,
                # and the player can reach the subsequent tile.
                if ((s.board[□ + game.step_fore[dir]] & (BOX+WALL+FLAG_BOX_REACHABLE_TILE)) == FLAG_BOX_REACHABLE_TILE) &&
                   (is_reachable(□ + 2*game.step_fore[dir], reach))
                    parking_spaces = find_greatest_distance_parking_candidates(
                                            packing_order.goal_and_parking_tiles[i_goal],
                                            game,
                                            s,
                                            distances_by_direction2,
                                            reach,
                                            stack,
                                            queue_big,
                                            tile_goal_distances
                                        )
                    if (is_valid(parking_spaces[1])) && # we found at least one
                       (
                          (parking_spaces[1].distance > best_parking_spaces[1].distance) ||
                          (
                            (parking_spaces[1].distance == best_parking_spaces[1].distance) &&
                            (!is_a_corner_goal_tile) # prefer non-corner-goals, as non-corner-goals tend to block more
                          )
                       )
                        best_parking_spaces = parking_spaces
                        best_i_goal = i_goal
                        break # exit the for loop
                    end
                end
            end
        end
    end

    return (best_i_goal, best_parking_spaces)
end


"""
The search is a recursive depth-first search.
Each successful call performs one of these operations:
1. removes the set of boxes that can be pulled to box starting positions, which haven't been used for that purpose earlier,
   or
2. moves one box from a target square (a goal square or a parking tile) to a parking tile
"""
# function  Search__(Depth__,SetCount__,SetMembersCount__,GoalAndParkingSquareCount__,RemainingBoxesCount__,RemainingRealGoalsCount__,PlayerPos__:Integer;
#                            SeenBoxStartPositions__:TBoxSetWithCount; {note that it's a value parameter, not a variable parameter}
#                            var ParkedTwiceGoalSet__:TBoxSetWithCount;
#                            var UsedBoxStartPositions__:TBoxIndexs):Boolean; {this is a depth-first recursive search where 'Depth__' is the search depth}
#         type TParkingSpace =record Distance,Square,PlayerPositionAfterParking:Integer; end;
#              TParkingSpaces=array[0..2] of TParkingSpace; {elements 1..2 are the true parking spaces; element 0 is only used as a local variable}
#         var  BoxNo,BoxStartPosIndex,GoalCandidate,GoalNo,i,Index,NewPlayerPos,OldGoalRemovalCandidatesCount,Square,
#              TotalReachableBoxStartingPositionsCount,
#              TotalReachableSquaresCount:Integer;
#              Direction:TDirection;
#              ParkingSpaces:TParkingSpaces;
#              GoalParkingCandidates,GoalRemovalCandidates:TBoxIndexSet;
#              GoalRemovalCandidatesBoxStartPositions:TBoxIndexs;
#              TestedGoalsSet:TGoalSet;
# 13393
function packing_order_search_!(
    data::PackingOrderSearchData,
    packing_order::PackingOrder,
    game::Game,
    s::State,  # NOTE: State must reflect game starting state, with boxes in starting positions
    □_player::TileIndex,
    reaches::Vector{ReachableTiles},
    stack::Vector{TileIndex},
    queue::Vector{Tuple{TileIndex, TileIndex, Int32}},
    distances_by_direction::Matrix{Int32}, # Of size board_size × N_DIRS
    reachable_box_starting_positions::Vector{BoxIndex},
    packing_order_type::PackingOrderType;
    depth::Int = 0,
    set_number::Int = 1, # The set number to which we add goals in this phase
    )::Bool

    println("-----------------------------------------------")
    println("Called packing_order_search_! with depth $depth")
    @show packing_order.goal_set_numbers[1:packing_order.goal_and_parking_tile_count]

    n_goals = length(game.goals)
    used_box_start_numbers = BoxIndex[]

    # TODO: We have to return this or something
    solver_push_count_increase = 0 # Keep track of the number of moves (either in pushes or macro box moves)



    # If true, all boxes have been removed from the board
    found_packing_order = data.remaining_boxes_count == 0
    if !found_packing_order
        # Not all boxes have been removed from the board
        println("$depth: Not all boxes have been removed from the board - conducting search")

        if set_number < length(reaches) - 2
            # The maximum search depth has not been exceeded.
            # (We use a different reach struct for each depth.)

            move_player!(s, □_player)
            reach = reaches[set_number+1] # (The first one is used by helper)
            calc_reachable_tiles!(reach, game, s.board, □_player, stack)

            goal_removal_candidates = BoxIndex[] # technically a set
            goal_removal_candidates_box_start_numbers = BoxIndex[] # technically a set
            goal_parking_candidates = BoxIndex[] # technically a set

            # Store this result from can_pull_box_to_starting_position for later use
            n_pulls = 0

            # The player may not be able to reach all boxes.
            # Add any boxes that are newly reachable from the current player position.
            if length(data.seen_box_start_positions) < length(s.boxes)
                for i_box in box_indices(game)
                    if (reach.tiles[s.boxes[i_box]] ≥ reach.calc_counter) &&
                        (i_box ∉ data.seen_box_start_positions)
                        push!(data.seen_box_start_positions, i_box)
                    end
                end
            end
            @show data.seen_box_start_positions

            if (length(data.seen_box_start_positions) == length(s.boxes)) || packing_order_type == PACKING_ORDER_PULL_BOXES
                # Boxes may be removed from the board when the player has access to all box starting positions
                # or PACKING_ORDER_PULL_BOXES means we simply pull boxes away from the goal squares.
                println("$depth: Player has access to all box starting positions!")

                # For each box on a goal tile, and for each parked box, check if the box can be removed from the board.
                # Boxes can be removed if ???
                # Any such boxes are added to goal_removal_candidates
                for i_goal in goal_and_parking_numbers(packing_order)

                    □_goal = packing_order.goal_and_parking_tiles[i_goal]
                    @show (i_goal, □_goal)

                    # Real goals have goal numbers equal to actual goals.
                    # Parking squares have larger goal numbers
                    is_parking_tile = i_goal > n_goals

                    set_number_is_unassigned = packing_order.goal_set_numbers[i_goal] == 0
                    goal_has_box = ((s.board[□_goal] & BOX) > 0)

                    if set_number_is_unassigned && goal_has_box

                        # Check to see whether it is possible to pull the box to a starting position.
                        can_pull_box_to_start_result = can_pull_box_to_starting_position(
                                i_goal,
                                game,
                                s,
                                reach,
                                stack,
                                queue,
                                distances_by_direction,
                                reachable_box_starting_positions,
                                packing_order.goal_and_parking_tiles,
                                packing_order.goal_and_parking_tile_count,
                                packing_order.goal_box_sets,
                                packing_order.goal_set_numbers,
                                packing_order.set_members)
                        n_pulls = can_pull_box_to_start_result.n_pulls

                        if can_pull_box_to_start_result.success || packing_order_type == PACKING_ORDER_PULL_BOXES

                            println("\t goal number $(i_goal) passed can_pull_box_to_starting_position!")

                            # Used to ensure we wait until all boxes not at box starting positions have been removed
                            box_on_start_square = ((s.board[□_goal] & FLAG_BOX_START_TILE) > 0)
                            all_boxes_at_starts = all_boxes_at_starting_positions(packing_order, s)

                            @show box_on_start_square
                            @show all_boxes_at_starts

                            # Only loop as long as the goal has not been registered as a candidate
                            goal_registered_as_candidate = false

                            for dir in DIRECTIONS
                                □_box_dst = □_goal + game.step_fore[dir]
                                □_player_dst = □_goal + 2game.step_fore[dir]

                                box_can_move_in_pull_dir = ((s.board[□_box_dst] & (BOX+WALL+FLAG_BOX_REACHABLE_TILE)) == FLAG_BOX_REACHABLE_TILE)

                                # NOTE: We could avoid the is_in_bounds check if we inlined and did <box_can_move_in_pull_dir> && <player_can_move_in_pull_dir>
                                player_can_move_in_pull_dir = is_in_bounds(□_player_dst, length(s.board)) && is_reachable(□_player_dst, reach)

                                pull_is_legal = box_can_move_in_pull_dir && player_can_move_in_pull_dir

                                @show (dir_to_char(dir), box_can_move_in_pull_dir, player_can_move_in_pull_dir)
                                if !goal_registered_as_candidate && pull_is_legal || (box_on_start_square && all_boxes_at_starts)

                                    if (can_pull_box_to_start_result.n_reachable_tiles ≥ data.large_board_area_threshold) || # the box comes from a relatively large area of the board; assume there will be a box in the area at the time the goal must be filled
                                       # (data.remaining_boxes_count ≤ end_game_phase_box_count_limit) || # most of the boxes, including parked boxes, have been removed from the board
                                       (packing_order_type == PACKING_ORDER_PULL_BOXES) ||
                                       (is_parking_tile) # it is a box from a parking tile as opposed to a box from a real goal square
                                        println("\tdir $(dir_to_char(dir)) is good")
                                        # It is often better to push boxes to parking tiles before pushing boxes to final goal squares
                                        # thus, boxes at goal squares and parking tiles are not removed from the board in the same game phase,
                                        # i.e., they are not getting the same packing order number
                                        latest_goal_removal_candidate = get(goal_removal_candidates, length(goal_removal_candidates), zero(BoxIndex))
                                        goal_candidate = i_goal
                                        if (!is_parking_tile) && (latest_goal_removal_candidate > n_goals)
                                            # Drop already found boxes from parking tiles that were candidates
                                            # for removal in this game phase.
                                            clear_removal_candidate_set!(
                                                goal_removal_candidates,
                                                goal_removal_candidates_box_start_numbers,
                                                packing_order,
                                                data,
                                                packing_order_type,
                                                game,
                                                s)
                                            goal_candidate = i_goal
                                        elseif is_parking_tile && (latest_goal_removal_candidate ≤ n_goals) && (!isempty(goal_removal_candidates) > 0)
                                            # Don't remove the parked box in this game phase; boxes are ready for removal directly from goal squares
                                            # Don't use the box from the parking tile as candidate for this game phase.
                                            goal_candidate = zero(BoxIndex)
                                        end

                                        if goal_candidate > 0
                                            println("\tAdding goal removal candidate! ", goal_candidate)
                                            add_goal_removal_candidate!(
                                                goal_removal_candidates,
                                                goal_removal_candidates_box_start_numbers,
                                                used_box_start_numbers,
                                                goal_candidate,
                                                packing_order,
                                                data,
                                                packing_order_type,
                                                can_pull_box_to_start_result.box_start_number,
                                                game,
                                                s,
                                                set_number)
                                            # Don't add the goal again in this dir loop
                                            goal_registered_as_candidate = true
                                        end
                                    else
                                        # The box comes from a relative small area of the board
                                        push!(goal_parking_candidates, i_goal)
                                        # Don't add the goal again in this dir loop
                                        goal_registered_as_candidate = true
                                    end
                                end
                            end # end for over directions
                        end
                    end
                end
            end

            println("\n$depth: Done computing candidates")
            println("\tgoal_parking_candidates: ", goal_parking_candidates)
            println("\tgoal_removal_candidates: ", goal_removal_candidates)

            # We have not yet found a full packing order
            found_packing_order = false

            println("$depth: Check to clear goal parking candidates")
            # It is often better to push boxes to parking tiles before pushing boxes to final goal squares.
            # Thus, boxes at goal squares and parking tiles are not removed from the board in the same
            # game phase, i.e., they are not getting the same packing order number
            latest_goal_removal_candidate = get(goal_removal_candidates, length(goal_removal_candidates), 0)
            boxes_from_goal_squares_ready_for_parking = !isempty(goal_parking_candidates) && goal_parking_candidates[1] ≤ n_goals
            if (latest_goal_removal_candidate ≤ n_goals) && # boxes from goal squares are ready for removal from the board in this game phase
               (boxes_from_goal_squares_ready_for_parking)
               println("\t clearing goal parking candidates")
                empty!(goal_parking_candidates)
            end

            println("$depth: Check to clear removal candidate set")
            if (latest_goal_removal_candidate > n_goals) && # the only boxes ready for removal come from parking tiles
               (data.remaining_real_goals_count > 0) && # some boxes haven't left the real goal squares yet
               (isempty(goal_parking_candidates))
                # It is often better not to remove parked boxes from the board before all the real goal squares have been emptied
                println("\t clearing goal removal candidate set")
                clear_removal_candidate_set!(
                    goal_removal_candidates,
                    goal_removal_candidates_box_start_numbers,
                    packing_order,
                    data,
                    packing_order_type,
                    game,
                    s)
            end

            # -------------------------------------------------------------------------------------
            # Check goal candidates ready for parking; park boxes.
            println("$depth: Check goal candidates ready for parking and park boxes")
            if (
                (!isempty(goal_parking_candidates)) || # boxes have been selected for parking because a goal square must be filled with a box coming from a relatively small area of the board
                (
                 isempty(goal_removal_candidates) &&
                 (packing_order_type == PACKING_ORDER_PULL_BOXES_TO_STARTING_POSITIONS_WITH_PARKING)
                )
               ) &&
               (packing_order.goal_and_parking_tile_count < length(packing_order.goal_and_parking_tiles)) # some goal numbers are still free to use as pseudo-goal numbers for parking tiles

                println("\t goal candidates are ready for parking!")

                # Remember the number of candidates ready for removal in this game phase
                old_goal_removal_candidates_count = length(goal_removal_candidates)

                # Clear tentative assignments for the boxes ready for removal in this game phase
                clear_removal_candidate_set!(
                    goal_removal_candidates,
                    goal_removal_candidates_box_start_numbers,
                    packing_order,
                    data,
                    packing_order_type,
                    game,
                    s)

                tested_goals_set = BoxIndex[] # technically a set
                if !isempty(goal_parking_candidates)
                    # one of the goal candidates should be pulled away now and parked in a relatively small reachable area of the board
                    for i_goal in goal_and_parking_numbers(packing_order)
                        if i_goal ∉ goal_parking_candidates
                            push!(tested_goals_set, i_goal)
                        end
                    end
                end

                # TODO: Ensure `reach` has been calculated.
                found_parking_candidates = true
                while (!found_packing_order && found_parking_candidates) # no full packing order has been found yet and we found a parking candidate

                    # Find good parking spaces
                    i_goal, parking_spaces = find_goal_candidates_for_parking(
                                                        game,
                                                        s,
                                                        distances_by_direction,
                                                        reach,
                                                        stack,
                                                        queue,
                                                        packing_order,
                                                        data,
                                                        tested_goals_set)
                    @show i_goal, parking_spaces
                    found_parking_candidates = is_valid(parking_spaces[1])
                    if found_parking_candidates
                        # 1 or 2 parking tiles have been found for the box at goal number 'GoalNo'
                        push!(tested_goals_set, i_goal) # each goal is only given one chance for parking
                        println("\t Setting goal number $(i_goal) to set number $(set_number)")
                        packing_order.goal_set_numbers[i_goal] = set_number # tentatively assign a packing order set number to the goal
                        packing_order.set_members[data.set_members_count+1] = i_goal # temporarily use 'set_members' to collect the goals and parking tiles (pseudo goals) in the packing order

                        # Remove the box from the goal square before it's moved to the parking tile
                        s.board[packing_order.goal_and_parking_tiles[i_goal]] &= ~BOX

                        # TODO: We disabled double-parking in find_goal_candidates_for_parking
                        # if i_goal > n_goals
                        #     # a parking tile as opposed to a real goal square
                        #     count += 1
                        #     push!(box_set, i_goal)
                        # else
                        data.remaining_real_goals_count -= 1
                        # end

                        for parking_space in parking_spaces
                            if !found_packing_order &&
                               is_valid(parking_space) &&
                               packing_order.goal_and_parking_tile_count < length(packing_order.goal_and_parking_tiles)

                                # place a box at the parking space
                                s.board[parking_space.□] |= BOX

                                # store the parking tile
                                packing_order.goal_and_parking_tile_count += 1
                                packing_order.goal_and_parking_tiles[packing_order.goal_and_parking_tile_count] = parking_space.□

                                # the parked box hasn't been assigned a packing order number yet
                                packing_order.goal_set_numbers[packing_order.goal_and_parking_tile_count] = 0

                                # the parked box hasn't been removed from the board yet, i.e.
                                # it hasn't been assigned a box starting position
                                push!(used_box_start_numbers, zero(BoxIndex))

                                # a parked box should preferably stay at its parking tile until
                                # it is time to fill the goal the parked box originally came from
                                packing_order.parked_box_destination_goal_set_numbers[packing_order.goal_and_parking_tile_count] = set_number

                                # update the number of pushes
                                solver_push_count_increase += parking_space.distance

                                # collect goal parking tiles
                                push!(data.goal_parking_tiles, i_goal)

                                # Calc the new reachable box starting positions for the parking square
                                empty!(packing_order.goal_box_sets[packing_order.goal_and_parking_tile_count])

                                # Recurse in the search with the newly added parking space
                                data.set_members_count += 1
                                data.remaining_boxes_count -= 0 # no-op

                                found_packing_order = packing_order_search_!(
                                    data,
                                    packing_order,
                                    game,
                                    s,
                                    □_player,
                                    reaches,
                                    stack,
                                    queue,
                                    distances_by_direction,
                                    reachable_box_starting_positions,
                                    packing_order_type,
                                    depth = depth+1,
                                    set_number = set_number+1)

                                data.set_members_count -= 1
                                data.remaining_boxes_count += 0 # no-op

                                # remove the parked box from the board again
                                s.board[parking_space.□] &= ~BOX

                                # check if new best found partial packing order,
                                # but with no boxes removed from the board yet.
                                if (!found_packing_order) &&
                                   (data.set_members_count + 1 > data.best_goal_set_count) &&
                                   (data.best_goal_set_remaining_boxes_count == length(game.goals))
                                    data.best_goal_set_count = data.set_members_count + 1
                                    data.best_goal_set_numbers = goal_set_numbers
                                end
                            end
                        end

                        # TODO: We disabled double-parking in find_goal_candidates_for_parking
                        # if i_goal > n_goals
                        #     # a parking tile as opposed to a real goal square
                        #     remove!(parked_twice_goal_set, i_goal)
                        # else
                            data.remaining_real_goals_count += 1
                        # end

                        # if   GoalNo>GoalCount then with ParkedTwiceGoalSet__ do begin {'True': a parking tile as opposed to a real goal square}
                        #      Dec(Count); Exclude(BoxSet,GoalNo);
                        #      end
                        # else Inc(RemainingRealGoalsCount__);

                        if found_packing_order # a packing order has been calculated
                            if length(goal_parking_candidates) > 0
                                # the parked box comes from a relatively small area of the board
                                println("The parked box comes from a relatively small area of the board!")
                                packing_order.last_packing_set_with_a_box_parking_in_a_small_area = max(
                                    set_number,
                                    packing_order.last_packing_set_with_a_box_parking_in_a_small_area
                                )
                            end
                        else
                            # the recursive call could not remove the remaining boxes from the board}
                            packing_order.goal_set_numbers[i_goal] = 0 # reset the packing order set number for the goal
                            data.goal_parking_tiles[i_goal] = 0 # reset the parking square for the goal
                            packing_order.parked_box_destination_goal_set_numbers[packing_order.goal_and_parking_tile_count] = 0
                            packing_order.last_packing_set_with_a_box_parking_in_a_small_area = 0 #  reset
                            packing_order.goal_and_parking_tile_count -= 1
                        end

                        # if Result then begin                                 {'True': a packing order has been calculated}
                        #    if GoalParkingCandidates.Count>0 then             {'True': the parked box comes from a relatively small area of the board}
                        #       LastPhaseWithABoxParkedInASmallArea:=Max(SetCount__,LastPhaseWithABoxParkedInASmallArea)
                        #    end
                        # else begin                                           {the recursive call could not remove the remaining boxes from the board}
                        #    GoalSetNo[GoalNo]:=0;                             {reset the packing order set number for the goal}
                        #    GoalParkingSquares[GoalNo]:=0;                    {reset the parking tile for the goal}
                        #    ParkedBoxDestinationGoalSetNo[Succ(GoalAndParkingSquareCount__)]:=0;
                        #    LastPhaseWithABoxParkedInASmallArea:=0;           {reset}
                        #    end;

                        s.board[packing_order.goal_and_parking_tiles[i_goal]] |= BOX # put the box back on the goal square
                        move_player!(s, □_player) # restore the player position
                    end
                end

                if !found_packing_order
                    # put goals ready for removal in this game phase back in the candidate-set
                    println("\t Did not find packing order - placing goals back into removal set")
                    for i in 1 : old_goal_removal_candidates_count
                        add_goal_removal_candidate!(
                                                goal_removal_candidates,
                                                goal_removal_candidates_box_start_numbers,
                                                used_box_start_numbers,
                                                goal_removal_candidates[i],
                                                packing_order,
                                                data,
                                                packing_order_type,
                                                can_pull_box_to_start_result.box_start_number,
                                                game,
                                                s,
                                                set_number)
                    end
                end
            end

            # -------------------------------------------------------------------------------------
            println("$depth: Check goal candidates ready for removal and remove boxes")
            # Check goal candidates ready for removal from the board; remove boxes
            # We do this if we have goal removal candidates and have not yet found a packing order.
            if !isempty(goal_removal_candidates) && !found_packing_order

                # Remove the goal removal candidates,
                # which means pull them from their goals to their starting tiles.
                for (i, i_goal) in enumerate(goal_removal_candidates)
                    packing_order.set_members[data.set_members_count + i] = i_goal

                    # Keep track of the number of pulls
                    if packing_order_type == PACKING_ORDER_PULL_BOXES_TO_STARTING_POSITIONS_WITHOUT_PARKING
                        solver_push_count_increase += n_pulls # TODO: Just using the latest value seems wrong
                    else
                        solver_push_count_increase += 1
                    end

                    # Assert that it is, in fact, a box
                    @assert (s.board[packing_order.goal_and_parking_tiles[i_goal]] & BOX) > 0

                    # Remove the box from the goal
                    s.board[packing_order.goal_and_parking_tiles[i_goal]] &= ~BOX
                end

                # Try to remove any remaining boxes from the board by recursively calling
                # this search method.
                n_goal_removal_candidates = length(goal_removal_candidates)
                data.set_members_count += n_goal_removal_candidates
                data.remaining_boxes_count -= n_goal_removal_candidates

                found_packing_order = packing_order_search_!(
                    data,
                    packing_order,
                    game,
                    s,
                    □_player,
                    reaches,
                    stack,
                    queue,
                    distances_by_direction,
                    reachable_box_starting_positions,
                    packing_order_type,
                    depth = depth+1,
                    set_number = set_number+1)

                data.set_members_count -= n_goal_removal_candidates
                data.remaining_boxes_count += n_goal_removal_candidates

                # if (not Result)
                #    and
                #    (BestGoalSetCount<=MAX_BOX_COUNT) {'True': store best partial order}
                #    and
                #    (((SetMembersCount__+GoalRemovalCandidates.Count)>BestGoalSetCount)
                #     or
                #     (BestGoalSetRemainingBoxesCount=Game.BoxCount)
                #    ) {'True': new best found partial order with boxes removed from the board}
                #    then begin
                #    BestGoalSetCount:=SetMembersCount__+GoalRemovalCandidates.Count;
                #    BestGoalSetNumbers:=GoalSetNo;
                #    BestGoalSetRemainingBoxesCount:=RemainingBoxesCount__-GoalRemovalCandidates.Count;
                #    end;

                # Keep track of the best partial order with boxes removed from the board,
                if (!found_packing_order) &&
                   (data.best_goal_set_count ≤ MAX_BOX_COUNT) && # store best partial order
                   (((data.set_members_count + n_goal_removal_candidates) > data.best_goal_set_count)
                    ||
                    (data.best_goal_set_remaining_boxes_count == n_goals)
                   )
                    data.best_goal_set_count = data.set_members_count + n_goal_removal_candidates
                    data.best_goal_set_remaining_boxes_count = data.remaining_boxes_count - n_goal_removal_candidates
                    copy!(data.best_goal_set_numbers, packing_order.goal_set_numbers)
                end

                # Put the removed boxes back on the board
                for i_goal in goal_removal_candidates
                    s.board[packing_order.goal_and_parking_tiles[i_goal]] |= BOX
                end

                if !found_packing_order
                    println("\t Did not find packing order - clear the removal candidate set")
                    clear_removal_candidate_set!(
                        goal_removal_candidates,
                        goal_removal_candidates_box_start_numbers,
                        packing_order,
                        data,
                        packing_order_type,
                        game,
                        s)
                end

            else
                println("\t Did not process goal removal candidates")
                # We did not process any goal removal candidates.
                # No boxes could be removed from the board.
                # Try to find boxes to park somewhere.
                if (depth == 0) && found_packing_order
                    # this is the top level call to the recursive search, and a packing order has been found
                    # perform post-processing of the found packing order, e.g., insert additional parking operations
                    post_process_packing_order()
                end
            end
        end
    else
        println("$depth: All boxes have been removed from the board!")
        println("$depth: Updating our packing order.")
        packing_order.set_count = set_number - 1
        # NOTE: This was originally saved, as goal_and_parking_tile_count was passed in the
        #       same way set_count is.
        # packing_order.goal_and_parking_tile_count = goal_and_parking_tile_count
    end

    return found_packing_order
end

# 13729
"""
Preconditions: All boxes have been removed from the board, and the player position isn't a goal square
Postcondition: All boxes have been removed from the board (from the goal positions and any parking tiles)
"""
function packing_order_search!(
    packing_order::PackingOrder,
    game::Game,
    s::State,  # NOTE: State must reflect game starting state, with boxes in starting positions
    □_player::TileIndex,
    reaches::Vector{ReachableTiles},
    stack::Vector{TileIndex},queue::Vector{Tuple{TileIndex, TileIndex, Int32}},
    distances_by_direction::Matrix{Int32}, # Of size board_size × N_DIRS
    reachable_box_starting_positions::Vector{BoxIndex},
    packing_order_type::PackingOrderType,
    large_board_area_threshold::Int,
    best_goal_set_count::Int,
    best_goal_set_remaining_boxes_count::Int
    )::Bool

    # Set the player's location
    move_player!(s, □_player)

    # Put boxes at the goals
    for □ in game.goals
        s.board[□] |= BOX
    end

    # Reset packing order memory
    copyto!(packing_order.goal_and_parking_tiles, game.goals) # First ones are the goals
    packing_order.goal_and_parking_tile_count = length(game.goals) # We initially only have the goals
    packing_order.set_count = 0
    packing_order.last_packing_set_with_a_box_parking_in_a_small_area = 0
    fill!(packing_order.goal_set_numbers, zero(eltype(packing_order.goal_set_numbers)))
    fill!(packing_order.parked_box_destination_goal_set_numbers, zero(eltype(packing_order.parked_box_destination_goal_set_numbers)))

    # Initialize some memory
    goal_parking_tiles = zeros(TileIndex, length(s.boxes))
    parked_twice_goal_set = BoxIndex[]

    # Boxes at goal squares are counted as `seen box starting positions`.
    seen_box_start_positions = TileIndex[] # technically a set
    for i_box in box_indices(game)
        □ = s.boxes[i_box]
        if (s.board[□] & GOAL) > 0
            push!(seen_box_start_positions, i_box)
        end
    end

    data = PackingOrderSearchData(
        0, # count
        best_goal_set_count,
        best_goal_set_remaining_boxes_count,
        similar(packing_order.goal_set_numbers), # best_goal_set_numbers
        Array{BoxIndex}(undef, length(game.goals)), # goal_group_numbers
        0, # set_members_count
        length(s.boxes), # remaining_boxes_count
        length(s.boxes), # remaining_real_goals_count
        goal_parking_tiles,
        parked_twice_goal_set,
        seen_box_start_positions,
        large_board_area_threshold,
        MAX_PACKING_ORDER_PARK_BOXES_ATTEMPTS)

    # if (packing_order_type == PACKING_ORDER_PULL_BOXES) && (0 < data.best_goal_set_count ≤ length(s.boxes))
        # Use the best found partial packing order from the previous run
        #    for GoalNo:=1 to GoalCount do                                      {find highest packing order set number}
        #        data.count:=Max(Count,BestGoalSetNumbers[GoalNo]);
        #    for Index:=1 to Count do begin                                     {for each set number}
        #        for GoalNo:=1 to GoalCount do begin                            {'GoalCount': only use the best found partial packing order for real goal squares, i.e., ignore parking tiles, if any}
        #            SetNo:=BestGoalSetNumbers[GoalNo];
        #            if SetNo=Index then with Solver.PackingOrder do begin
        #               data.set_members_count+=1
        #               Dec(RemainingBoxesCount);
        #               SetMembers[data.set_members_count]:=GoalNo;
        #               GoalSetNo[GoalNo]:=Index;
        #               Board[GoalPos[GoalNo]]:=Board[GoalPos[GoalNo]] and (not BOX); {remove boxes from goal squares which already have been assigned a packing order group number}
        #               if GoalNo<=GoalCount then
        #                  Dec(RemainingRealGoalsCount);
        #               end;
        #           end;
        #        end;
    # end

    found_packing_order = packing_order_search_!(
        data,
        packing_order,
        game,
        s,
        s.player,
        reaches,
        stack,
        queue,
        distances_by_direction,
        reachable_box_starting_positions,
        packing_order_type)

    # Remove boxes from the goals
    for □ in game.goals
        s.board[□] &= ~BOX
    end

    # Restore player position
    move_player!(s, □_player)

    return found_packing_order
end

# 13779
function calculate_packing_order_helper!(
    packing_order::PackingOrder,
    game::Game,
    s::State,
    reaches::Vector{ReachableTiles},
    stack::Vector{TileIndex},
    queue::Vector{Tuple{TileIndex, TileIndex, Int32}},
    distances_by_direction::Matrix{Int32}, # Of size board_size × N_DIRS
    reachable_box_starting_positions::Vector{BoxIndex},
    n_box_reachable_tiles::Int,
    packing_order_type::PackingOrderType
    )

    # Clear the reachable tiles so that a visited square contains 0.
    clear!(reaches[1], s.board)

    # Save original player position
    □_player_orig = s.player

    # Remove all boxes from the board and clear the visited flag for all box starting positions
    for □ in s.boxes
        s.board[□] &= ~(BOX + FLAG_VISITED_TILE)
    end

    end_game_phase_box_count_limit = div(length(s.boxes)*(100-END_GAME_PHASE_PERCENT)+50, 100)
    large_board_area_threshold = div(n_box_reachable_tiles*LARGE_BOARD_AREA_THRESHOLD_PERCENT, 100)

    best_goal_set_count = 0
    if packing_order_type == PACKING_ORDER_PULL_BOXES_TO_STARTING_POSITIONS_WITHOUT_PARKING
        calculate_tunnel_squares!() # calculate the tunnel squares so parking at 'do not stop here' squares can be avoided
        best_goal_set_count = length(s.boxes)+1 # disable saving of the best found partial packing order
    elseif packing_order_type == PACKING_ORDER_PULL_BOXES_TO_STARTING_POSITIONS_WITH_PARKING && best_goal_set_count > length(s.boxes)
        # Enable saving of the best found partial packing order,
        # so it is available for a plain "pull boxes away from goal" search if the search with parking fails.
        best_goal_set_count = 0
    end

    # Store the smallest number of remaining boxes.
    # This is initialized to the total number of boxes, as we have not removed any yet.
    best_goal_set_remaining_boxes_count = length(s.boxes)

    # Run until a packing order has been found, or until all player access areas on
    # the board have been tested.
    first_time = true
    found_packing_order = false
    while first_time || !(found_packing_order || s.player == □_player_orig)
        first_time = false
        @show s.player

        #########################################################################################
        # Get reachability with boxes at all of the goals

        # Put boxes at the goals
        for □ in game.goals
            s.board[□] |= BOX
        end

        # Grab the player position.
        # If the player is on a goal tile, move them to any other reachable position.
        □ = s.player
        while ((s.board[□] & (BOX+WALL+FLAG_BOX_REACHABLE_TILE)) != FLAG_BOX_REACHABLE_TILE) &&
              (reaches[1].tiles[□] != 0) &&
              (□ != zero(TileIndex))
            # Try the next player tile
            □ = TileIndex(mod1(□ + one(TileIndex), length(s.board)))
            if □ == s.player
                # All squares have been visited without finding an empty box-reachable time
                # next to a tile square. The level is probably malformed.
                □ = zero(TileIndex)
            end
        end
        move_player!(s, □)

        # The reachable tile timestamps were cleared before the while loop,
        # so the tiles in an unvisited player access area are 0.
        calc_reachable_tiles!(reaches[1], game, s.board, □, stack)

        # Remove boxes from the goals
        # TODO: This may be redundant with put-boxes-at-goals at the beginning of packing_order_search!
        for □ in game.goals
            s.board[□] &= ~BOX
        end

        #########################################################################################
        # Now run a search
        # Note that this call will remove boxes from the board (from goal squares and any parking tiles)
        found_packing_order = packing_order_search!(
            packing_order,
            game,
            s,
            s.player,
            reaches,
            stack,
            queue,
            distances_by_direction,
            reachable_box_starting_positions,
            packing_order_type,
            large_board_area_threshold,
            best_goal_set_count,
            best_goal_set_remaining_boxes_count)

        if !found_packing_order
            # The goal squares could not be emptied when the player started from their
            # current starting tile. Advance to the next empty reachable floor square.
            □ = s.player
            □ = TileIndex(mod1(□ + one(TileIndex), length(s.board)))
            @show □
            while !((
                    (reaches[1].tiles[□] == 0) &&
                    ((s.board[□] & (WALL+GOAL+FLOOR)) == FLOOR)
                  ) || (□ == □_player_orig)) # Or we loop back to our original position
                □ = TileIndex(mod1(□ + one(TileIndex), length(s.board)))
            end
            @show "after", □
            move_player!(s, □)
        end
    end

    place_all_boxes_on_board!(game, s)
    move_player!(s, □_player_orig)

    # 13843
    if found_packing_order
        println("FOUND A PACKING ORDER!!")

        if packing_order.goal_and_parking_tile_count > length(game.goals)
            # Some boxes require parking.
            # Calculate the distances from all squares to the parking squares.
            # The parking squares are handled as pseudo-goal squares.
            @assert false # TODO
            calc_tile_goal_distances_and_goal_box_sets!(
                tile_goal_distances::Matrix{Int32}, # of size n_tiles × n_goals
                goal_box_sets::Vector{Vector{UInt8}},
                goal_set_numbers::Vector{UInt8}, # mapping of each goal number (or parking square) to its packing order set number - ie its phase
                parked_box_destination_goal_set_numbers::Vector{UInt8}, # for each parked box, the set number of the goal the parked box came from; a parked box should preferably stay at its parking square until it's time to fill the goal the parked box originally came from
                BoxIndex(length(game.goals)),
                packing_order.goal_and_parking_tile_count,
                destination_offset::Int,
                game::Game,
                s::State,
                freeze_data::ReachableTiles,
                distances_by_direction::Matrix{Int32},
                reach::ReachableTiles,
                stack::Vector{TileIndex},
                queue::Vector{Tuple{TileIndex, TileIndex, Int32}}
                )
        end

        #= TODO
        if ( 2 * GoalAndParkingSquareCount ) < High( TBoxNo ) then begin
            # the square->goal distance table is big enough to contain the
            # distances calculated taking the packing order into account;
            # calculate these more accurate distances and use them during the
            # search when the game state evaluation function calculates bonuses
            # and penalties based on distances to goals and parking spaces to
            # fill in the current game phase
            DistancesBasedOnPackingOrder:=GoalAndParkingSquareCount; {offset from the normal distances where the packing order aware distances can be found}
            CalculateSquareGoalDistances(1,GoalAndParkingSquareCount,DistancesBasedOnPackingOrder,True,SquareGoalDistance,GoalBoxSets); {calculate distances for real goal squares and for parking squares}
        end
        =#

        @show packing_order.set_count
        @show packing_order.goal_set_numbers[1:packing_order.goal_and_parking_tile_count]

        # Count the number of goals in each set.
        # Note that `set_sizes` was originally an array.
        set_sizes = Dict{UInt8, Int}()
        for i_goal in goal_and_parking_numbers(packing_order)
            set_number = packing_order.goal_set_numbers[i_goal]
            set_sizes[set_number] = get(set_sizes, set_number, zero(BoxIndex)) + 1
        end

        # Compute packing_order.first_set_member_index, which is the
        # index of first goal member in `set_members` for each set number.
        # NOTE: We deviate from YASS by not adding the sentiel to the end of the array.
            # i:=1; FirstSetMemberIndex[0]:=0; SetSize[0]:=1;
            # while i<=SetCount do begin
            #   if SetSize[i]<>0 then begin
            #      FirstSetMemberIndex[i]:=FirstSetMemberIndex[Pred(i)]+SetSize[Pred(i)];
            #      Inc(i);
            #      end
            #   else begin {empty set; remove it}
            #      for GoalNo:=1 to GoalAndParkingSquareCount do begin
            #          if GoalSetNo                    [GoalNo]>i then Dec(GoalSetNo                    [GoalNo]);
            #          if ParkedBoxDestinationGoalSetNo[GoalNo]>i then Dec(ParkedBoxDestinationGoalSetNo[GoalNo]);
            #          end;
            #      Dec(SetCount);
            #      for j:=i to SetCount do SetSize[j]:=SetSize[Succ(j)];
            #      end;
            #   end;
            # FirstSetMemberIndex[Succ(SetCount)]:=FirstSetMemberIndex[SetCount]+SetSize[SetCount]; {the set number after the last set contains a sentinel, i.e., the first unused list member number}
        let
            packing_order.first_set_member_index[1] = 1 # First set starts at the first index
            i = 2
            while i ≤ packing_order.set_count
                if get(set_sizes, i, zero(BoxIndex)) > 0
                    packing_order.first_set_member_index[i] =
                        packing_order.first_set_member_index[i-1] + get(set_sizes, i, 0)
                    i += 1
                else
                    # It is an empty set - delete it and decrement all following goal set numbers
                    for i_goal in goal_and_parking_numbers(packing_order)
                        if packing_order.goal_set_numbers[i_goal] > i
                            packing_order.goal_set_numbers[i_goal] -= 1
                        end
                        if packing_order.parked_box_destination_goal_set_numbers[i_goal] > i
                            packing_order.parked_box_destination_goal_set_numbers[i_goal] -= 1
                        end
                    end
                    packing_order.set_count -= 1
                    for j in i:packing_order.set_count
                        set_sizes[j] = get(set_sizes, j+1, 0)
                    end
                end
            end
            # Set the sentinel value
            packing_order.first_set_member_index[packing_order.set_count+1] =
                packing_order.first_set_member_index[packing_order.set_count] + set_sizes[packing_order.set_count]
        end

        @show set_sizes
        @show packing_order.first_set_member_index[1:packing_order.set_count+1]

        # Compute packing_order.set_members, ex:
        # sets A,B, and C with 2, 3, and 2 members look like: [a1,a2,b1,b2,b3,c1,c2]
        for i_goal in goal_and_parking_numbers(packing_order)
            set_number = packing_order.goal_set_numbers[i_goal]
            if set_number != 0
                set_sizes[set_number] -= 1
                i_first = packing_order.first_set_member_index[set_number]
                packing_order.set_members[i_first + set_sizes[set_number]] = i_goal
            end
        end
        @show packing_order.set_members[1:packing_order.set_count]

        # Compute the final (lowest) game phase where a box is parked because
        # a goal tile must be filled with a box coming from a relatively small area of the board.
        let
            i = packing_order.last_packing_set_with_a_box_parking_in_a_small_area

            # Number of boxes parked in a small area because a goal must be filled with a box
            # coming from a relatively small area of the board.
            packing_order.last_packing_set_with_a_box_parking_in_a_small_area = packing_order.set_count + 1

            for i_goal in goal_and_parking_numbers(packing_order)
                set_number = packing_order.parked_box_destination_goal_set_numbers[i_goal]
                if 0 < set_number ≤ i
                    packing_order.last_packing_set_with_a_box_parking_in_a_small_area = min(
                        packing_order.goal_set_numbers[i_goal],
                        packing_order.last_packing_set_with_a_box_parking_in_a_small_area
                    )
                end
            end
        end
        @show packing_order.last_packing_set_with_a_box_parking_in_a_small_area
    else
        # No packing order
        packing_order.set_count = 0
        packing_order.goal_and_parking_tile_count = length(game.goals)
        fill!(packing_order.goal_set_numbers, zero(eltype(packing_order.goal_set_numbers)))
        packing_order.last_packing_set_with_a_box_parking_in_a_small_area = 0
    end

    # ------------------------------------------------------------------------------------
    # Restore the board and store the goal numbers in the board square values
    # for all goal tiles and parking tiles.

    # Ensure that goals are back on the board
    for □_goal in game.goals
        s.board[□_goal] |= GOAL
    end

    # Remove any old goal numbers that link back from board tiles to goals
    for □ in tile_indices(game)
        s.board[□] &= ((1 << GOAL_BIT_SHIFT_COUNT)-1)
    end

    for i_goal in 1:packing_order.goal_and_parking_tile_count
        □ = game.goals[i_goal]
        v = s.board[□]
        # The existing goal number for this square.
        # A parking square may overlap real goal squares and previous parking squares
        □_i_goal = get_i_goal(v)
        # Remove temporary wall, if any, from the goal square
        s.board[□] = v & (~WALL)
        # Substract the previously inserted goal number from the square value
        s.board[□] -= (□_i_goal << GOAL_BIT_SHIFT_COUNT)
        # Insert the goal number in the square value, so there is a mapping
        # from the board square values to the highest goal/parking place numbers.
        s.board[□] += (i_goal << GOAL_BIT_SHIFT_COUNT)
    end

    # Initially, make a circular list for each goal, where the goal itself is the only member of the list
    for i_goal in 1:(packing_order.goal_and_parking_tile_count)
        packing_order.next_goal_at_same_square[i_goal] = i_goal
    end

    for i_goal in (packing_order.goal_and_parking_tile_count):-1:1
        v = s.board[game.goals[i_goal]]

        # The highest real goal number or parking space for this square.
        # A parking square may overlap real goal squares and other parking squares.
        □_i_goal = get_i_goal(v)

        # Link from the current goal to the next member of the circular list of goals for this square
        packing_order.next_goal_at_same_square[i_goal] = packing_order.next_goal_at_same_square[□_i_goal]

        # Link from the highest goal number for this square to the current goal number.
        # This is the link from the last member of the circular list back to the first member.
        packing_order.next_goal_at_same_square[□_i_goal] = i_goal
    end

    return found_packing_order
end


# 13991
# function calculate_packing_order(game::Game s::State)
#     t_start = time()

    # n = length(game.goals)
    # packing_order = PackingOrder(
    #     zero(Int32),
    #     zeros(Int32, n),
    #     [Set{UInt8}() for i in 1:n],
    #     zeros(UInt8, n),
    #     zero(Int32),
    #     zero(Int32),
    #     zeros(UInt8, n),
    #     zeros(UInt8, n),
    #     zero(Int32),
    #     zeros(UInt8, n),
    #     zeros(UInt8, length(game.board), n),
    #     zero(Float64),
    # )

    #     connected_goals_count, goal_group_count, goal_group_numbers = calculate_connected_goals(game)
#     calc_tunnel_squares!(game)

#     # Then try pulling all boxes to box starting positions with intermediate steps, parking boxes at suitable squares
#     succeeded = calculate_packing_order_helper(PACKING_ORDER_PULL_BOXES_TO_STARTING_POSITIONS_WITH_PARKING)
#     if !succeeded
#         # As a fallback method, simply pull boxes away from goal squares
#         calculate_packing_order_helper(PACKING_ORDER_PULL_BOXES)
#     end

#     packing_order.calc_time = time() - t_start
#     return packing_order
# end