"""
Evaluates the given list of pushes as .
"""
struct SolutionEval
    solved::Bool
    n_moves::Int
    n_pushes::Int
end

function evaluate_solution(
    game::Game,
    s0::State,
    pushes::Vector{Push},
    reach::ReachableTiles,
    stack::Vector{TileIndex},
    distances::Vector{Int32},
    queue::Vector{Tuple{TileIndex, Int32}},
    )
    n_moves = 0
    n_pushes = 0

    s = deepcopy(s0)
    clear!(reach, s.board)

    for a in pushes
        ▩ = s.boxes[a.i_box] # box location
        □ = ▩ - game.step_fore[a.dir] # player push start location
        ⬓ = ▩ + game.step_fore[a.dir] # box destination

        if ((s.board[□] & (WALL + BOX)) > 0)
            # we're trying to push from a wall or box
            return SolutionEval(false, n_moves, n_pushes)
        elseif ((s.board[⬓] & (WALL + BOX)) > 0)
            # we're trying to push into a wall or box
            return SolutionEval(false, n_moves, n_pushes)
        end

        calc_reachable_tiles!(reach, game, s.board, s.player, stack)

        # Get distance to target
        player_orig = s.player
        s.player = □
        calc_distances!(distances, game, s, reach, queue)
        s.player = player_orig

        # no path to target
        if distances[player_orig] == typemax(eltype(distances))
            return SolutionEval(false, n_moves, n_pushes)
        end

        while s.player != □
            # Find a neighbor square that is closer, if it exists
            step_dir = typemax(Direction)
            ⬓ = zero(TileIndex)
            for dir in DIRECTIONS
                ◳ = s.player + game.step_fore[dir]
                if ⬓ == 0 || distances[◳] < distances[⬓]
                    ⬓ = ◳
                    step_dir = dir
                end
            end
            if step_dir == typemax(Direction)
                return SolutionEval(false, n_moves, n_pushes)
            end

            n_moves += 1
            move!(s, game, step_dir)
        end

        # Now apply the push
        n_moves += 1
        n_pushes += 1
        move!(s, game, a)

        # check for a solution (despite pushes not being over)
        if is_solved(game, s)
            break
        end
    end

    return SolutionEval(true, n_moves, n_pushes)
end
function evaluate_solution(
    game::Game,
    s0::State,
    pushes::Vector{Push}
    )

    reach = ReachableTiles(length(s.board))
    clear!(reach, s.board)
    stack = Array{TileIndex}(undef, length(game.board))
    distances = zeros(Int32, length(s.board))
    queue = Array{Tuple{TileIndex, Int32}}(undef, length(s.board))

    return evaluate_solution(game, s0, pushes, reach, stack, distances, queue)
end

"""
Convert a trajectory in pushes from the given start state to a trajectory in moves.
The given state should have the player in their correct starting position.
NOTE: We allocate memory.
"""
function convert_to_moves(game::Game, s0::State, pushes::Vector{Push})
    # Allocate memory
    s = deepcopy(s0)
    reach = ReachableTiles(length(s.board))
    clear!(reach, s.board)
    stack = Array{TileIndex}(undef, length(game.board))
    distances = zeros(Int32, length(s.board))
    queue = Array{Tuple{TileIndex, Int32}}(undef, length(s.board))

    moves = Direction[]
    for a in pushes
        ▩ = s.boxes[a.i_box] # box location
        □ = ▩ - game.step_fore[a.dir] # player push start location

        calc_reachable_tiles!(reach, game, s.board, s.player, stack)

        # Get distance to target
        player_orig = s.player
        s.player = □
        calc_distances!(distances, game, s, reach, queue)
        s.player = player_orig


        while s.player != □
            # Find a neighbor square that is closer.
            # We assume that it exists
            step_dir = typemax(Direction)
            ⬓ = zero(TileIndex)
            for dir in DIRECTIONS
                ◳ = s.player + game.step_fore[dir]
                if ⬓ == 0 || distances[◳] < distances[⬓]
                    ⬓ = ◳
                    step_dir = dir
                end
            end
            @assert step_dir != typemax(Direction)

            push!(moves, step_dir)
            move!(s, game, step_dir)
        end

        # Now apply the push
        push!(moves, a.dir)
        move!(s, game, a)
    end

    return moves
end


"""
Takes a given push solution and attempts to get a better move-distance solution.
This implementation attempts to swap neighboring push pairs for different blocks
to reduce the total move distance without disrupting the solution.
"""
function optimize_push_solution!(game::Game, s0::State, pushes::Vector{Push})
    reach = ReachableTiles(length(s.board))
    clear!(reach, s.board)
    stack = Array{TileIndex}(undef, length(game.board))
    distances = zeros(Int32, length(s.board))
    queue = Array{Tuple{TileIndex, Int32}}(undef, length(s.board))

    eval = evaluate_solution(game, s0, pushes, reach, stack, distances, queue)
    if !eval.solved
        println("The push-solution you asked me to optimize is invalid")
        return (pushes, eval)
    end

    pushes′ = deepcopy(pushes)

    n_iters = 0
    n_swaps_neighbor = 0
    n_swaps_sequence = 0
    n_swaps_chain = 0
    improved  = true
    while improved
        n_iters += 1
        println(n_iters, ": ", eval.n_moves, "  ", n_swaps_neighbor, "  ", n_swaps_sequence, "  ", n_swaps_chain)
        improved = false

        for push in pushes
            print('a' + push.i_box - 1)
        end
        println()

        # Consider all neighbor box pairs
        for i in 2 : length(pushes)
            pa = pushes[i-1]
            pb = pushes[i]

            # Only consider neighboring pairs of pushes for different boxes
            if pa.i_box != pb.i_box
                # swap the push order
                pushes[i], pushes[i-1] = pushes[i-1], pushes[i]
                eval′ = evaluate_solution(game, s0, pushes, reach, stack, distances, queue)
                if eval′.solved && eval′.n_moves < eval.n_moves
                    improved = true
                    n_swaps_neighbor += 1
                    eval = eval′
                else
                    # undo
                    pushes[i], pushes[i-1] = pushes[i-1], pushes[i]
                end
            end
        end

        # Every time we change which box we are pushing,
        # consider moving all push sequences of all other next boxes
        # to before this sequence of box pushes.
        # ex: push      aaabbccbad
        #     consider: bbaaaccbad
        #               ccaaabbbad
        #               daaabbccba
        let
            a_lo = 1
            boxes_done = falses(length(s.boxes))
            while a_lo <= length(pushes)
                fill!(boxes_done, false)

                # a_lo:a_hi are all pushes with the same box
                pa = pushes[a_lo]
                boxes_done[pa.i_box] = true
                a_hi = a_lo
                while a_hi < length(pushes) && pushes[a_hi+1].i_box == pa.i_box
                    a_hi += 1
                end

                b_lo = a_hi + 1
                while b_lo <= length(pushes)
                    b_lo = findnext(p -> !boxes_done[p.i_box], pushes, b_lo)
                    if isnothing(b_lo)
                        b_lo = length(pushes) + 1
                        break
                    end

                    # b_lo:b_hi are all pushes with the next box
                    pb = pushes[b_lo]
                    boxes_done[pb.i_box] = true
                    b_hi = b_lo
                    while b_hi < length(pushes) && pushes[b_hi+1].i_box == pb.i_box
                        b_hi += 1
                    end

                    # now consider moving b_lo:b_hi up in front of a_lo:a_hi
                    # TODO: Consider moving subsequences
                    copy!(pushes′, pushes)
                    k = a_lo
                    for b in b_lo:b_hi
                        pushes′[k] = pushes[b]
                        k += 1
                    end
                    for c in a_hi+1 : b_lo-1
                        pushes′[k] = pushes[c]
                        k += 1
                    end
                    for a in a_lo:a_hi
                        pushes′[k] = pushes[a]
                        k += 1
                    end

                    eval′ = evaluate_solution(game, s0, pushes′, reach, stack, distances, queue)
                    if eval′.solved && eval′.n_moves < eval.n_moves
                        improved = true
                        n_swaps_sequence += 1
                        eval = eval′
                        copy!(pushes, pushes′)

                        # complete this iteration, as we've now swapped things
                        b_lo = length(pushes) + 1
                        break
                    end
                end

                a_lo = b_lo
            end
        end

        # For each push, if the next push is not for the same box,
        # and there is another sequence of pushes for this box afterwards,
        # try moving the largest possible sequence push earlier to immediately after.
        # i.e., try chaining pushes
        let
            a_lo = 1
            while a_lo <= length(pushes)

                # a_lo:a_hi are all pushes with the same box
                pa = pushes[a_lo]
                a_hi = a_lo
                while a_hi < length(pushes) && pushes[a_hi+1].i_box == pa.i_box
                    a_hi += 1
                end

                b_lo = findnext(p -> p.i_box == pa.i_box, pushes, a_hi+1)
                if isnothing(b_lo)
                    a_lo = a_hi + 1
                    continue
                end

                # b_lo:b_hi are all pushes with the next box
                b_hi = b_lo
                while b_hi < length(pushes) && pushes[b_hi+1].i_box == pa.i_box
                    b_hi += 1
                end

                # now consider moving b_lo:b_mid after a_hi
                for b_mid in b_hi : -1 : b_lo
                    copy!(pushes′, pushes)
                    for (bi, b) in enumerate(b_lo:b_mid)
                        pushes′[a_hi + bi] = pushes[b]
                    end
                    for (ci, c) in enumerate(a_hi+1:b_lo-1)
                        pushes′[a_hi + b_mid - b_lo + ci + 1] = pushes[c]
                    end

                    eval′ = evaluate_solution(game, s0, pushes′, reach, stack, distances, queue)
                    if eval′.solved && eval′.n_moves <= eval.n_moves
                        improved = true
                        n_swaps_chain += 1
                        eval = eval′
                        copy!(pushes, pushes′)

                        # complete this iteration, as we've now swapped things
                        a_hi = length(pushes)
                        break
                    end
                end

                a_lo = a_hi+1
            end
        end
    end

    return (pushes, eval)
end