# This method is not called directly, but is used inside `calculate_connected_goals`
function calculate_connected_goals_helper!(
    goal_group_numbers::Vector{UInt8}, # One for each goal
    game::Game,
    □_goal::TileIndex,
    visit_count::Int, # (starts off as 0)
    group_number::UInt8,
    )::Int

    n_goals_in_group = visit_count + 1

    # Assign the current group number to this goal
    goal_group_numbers[get_i_goal(game.board[□_goal])] = group_number
    for dir in DIRECTIONS
        □_neighbor = □_goal + game.step_fore[dir]
        if ((game.board[□_neighbor] & GOAL) > 0) && # The neighbor square is also a goal
           (goal_group_numbers[get_i_goal(game.board[□_neighbor])] == 0) && # the neighbor goal hasn't been assigned a group number yet
           (
            # The goal has an unvisited neighboring goal in this direction
            # but the squares along the other axis is also taken into
            # account; this improves the program's ability to distinguish
            # between true goal-rooms and goals that just happen to be
            # loosely connected, for instance by being on a line in the
            # middle of the board;
            # Only goals with a neighboring goal or a neighboring wall on the
            # other axis are counted as connected goals, counting toward the
            # threshold for the packing-order search;
              ((game.board[□_goal + game.step_left[dir]] & (WALL+GOAL)) > 0) ||
              ((game.board[□_goal + game.step_right[dir]] & (WALL+GOAL)) > 0)
            )

            n_goals_in_group = calculate_connected_goals_helper!(goal_group_numbers, game, □_neighbor, n_goals_in_group, group_number)
        end
    end

    return n_goals_in_group
end

# 12611
function calculate_connected_goals(game::Game)

    connected_goals_count = 0
    goal_group_count = zero(UInt8)
    goal_group_numbers = zeros(UInt8, length(game.goals)) # For each goal, the id of the group it belongs to

    # Find connected goal groups
    for (i_goal, □) in enumerate(game.goals)
        if goal_group_numbers[i_goal] == 0
            # The goal square has not been visited before,
            # meaning it hasn't been assigned a group number yet
            goal_group_count += one(UInt8)
            n_goals_in_group = calculate_connected_goals_helper!(goal_group_numbers, game, □, 0, goal_group_count)
            if n_goals_in_group > 1
                connected_goals_count += n_goals_in_group
            end
        end
    end

    return (connected_goals_count, goal_group_count, goal_group_numbers)
end

"""
Finds all door squares and all gate squares, setting FLAG_DOOR_TILE and FLAG_GATE_TILE in game.board.
A door has two walls on one axis, wheras a gate separates two rooms if a box is placed in it.
A gate square splits the board in separate rooms.
Returns the number of gate squares.
We expect FLAG_ILLEGAL_BOX_TILE and FLAG_BOX_REACHABLE_TILE to have already been computed.
"""
function calc_door_and_gate_squares!(
    game::Game,
    s::State, # start state
    reach::ReachableTiles,
    stack::Vector{TileIndex},
    )

    n_gate_squares = 0

    remove_all_boxes_from_board!(game, s)

    for □ in tile_indices(game)
        if (game.board[□] & (WALL+FLAG_ILLEGAL_BOX_TILE+FLAG_BOX_REACHABLE_TILE+FLOOR+FLAG_GATE_TILE)) == (FLAG_BOX_REACHABLE_TILE+FLOOR)
            # Is a legal and box-reachable tile and is not yet a gate tile
            for dir in DIRECTIONS
                ▩ = □ + game.step_fore[dir] # neighbor tile
                ⬓ = □ - game.step_fore[dir] # opposite neighbor tile

                if ((game.board[▩] & FLOOR) > 0) && ((game.board[⬓] & FLOOR) > 0)
                    if (game.board[□ + game.step_left[dir]] & WALL) > 0 &&
                       (game.board[□ + game.step_right[dir]] & WALL) > 0
                        # It is a door tile
                        game.board[□] |= FLAG_DOOR_TILE
                    end

                    if !(is_a_legal_and_box_reachable_tile(game.board[□+game.step_left[dir]]) ||
                         is_a_legal_and_box_reachable_tile(game.board[□+game.step_right[dir]]))

                        #  -W-  'W' = wall or illegal box tile;
                        #  >□>  direction ->
                        #  -W-

                        game.board[□] |= BOX
                        calc_reachable_tiles!(reach, game, game.board, ▩, stack)
                        game.board[□] &= ~BOX

                        if !are_connected(▩, ⬓, reach)
                            # 'Square' splits the board in 2 separate rooms
                            game.board[□] |= FLAG_GATE_TILE
                            n_gate_squares += 1
                        end
                    elseif (((game.board[□ + game.step_left[dir]] & WALL) > 0) &&
                            (((game.board[▩ + game.step_right[dir]] & WALL) > 0) ||
                             ((game.board[⬓ + game.step_right[dir]] & WALL) > 0))) ||
                            #  -#-       -#-    '#' = wall
                            #  >□>  or   >□>    direction ->
                            #  --#       #--
                           (((game.board[▩ + game.step_left[dir]] & WALL) > 0) &&
                            ((game.board[▩ + game.step_left[dir]] & WALL) > 0))
                            #  #-#    '#' = wall
                            #  >□>    direction ->
                            #  ---
                        game.board[□] |= BOX
                        calc_reachable_tiles!(reach, game, game.board, ▩, stack)
                        if !are_connected(▩, ⬓, reach)
                            # □ splits the board in 2 separate rooms when the box moves along the axis given by 'dir'
                            ▩ = □ + game.step_left[dir]
                            ⬓ = □ + game.step_right[dir]
                            # ex:,
                            #   #-#    '#' = wall; middle = ▩
                            #   >□>    direction ->
                            #   --#    middle = new ⬓
                            if ((game.board[▩] & FLOOR) > 0) && ((game.board[⬓] & FLOOR) > 0)
                                calc_reachable_tiles!(reach, game, game.board, ▩, stack)
                                if !are_connected(▩, ⬓, reach)
                                    # □ splits the board in 2 separate rooms
                                    game.board[□] |= FLAG_GATE_TILE
                                    n_gate_squares += 1
                                end
                            else
                                # ▩ or ⬓ contains a wall, ex:
                                #  ##-    '#' = wall; middle = ▩
                                #  >□>    direction ->
                                #  --#    middle = ⬓
                                game.board[□] |= FLAG_GATE_TILE
                                n_gate_squares += 1
                            end
                        end
                        game.board[□] &= ~BOX
                    end
                end
            end
        end
    end

    place_all_boxes_on_board!(game, s)

    return n_gate_squares
end


"""
Compute all tunnel squares, setting the tunnel square flags.
Note, we expect FLAG_BOX_REACHABLE_TILE, FLAG_ILLEGAL_BOX_TILE, and FLAG_GATE_TILE to be set.
"""
function calc_tunnel_squares!(game::Game)
    # Clear old flags, if any
    for □ in tile_indices(game)
        game.board[□] &= ~DIRECTION_TO_TUNNEL_FLAGS
    end
    for □ in tile_indices(game)
        if is_a_legal_and_box_reachable_tile(game.board[□])
            for dir in DIRECTIONS
                ▩ = □ + game.step_fore[dir]
                if is_a_legal_and_box_reachable_tile(game.board[▩])
                    tile_is_blocked_by_two_walls_along_another_axis =
                        ((game.board[□ + game.step_left[dir]] & WALL) > 0) &&
                        ((game.board[□ + game.step_right[dir]] & WALL) > 0)

                    # A box at □ can only move along the axis given by
                    # dir, and both tiles along the other axis
                    # contain walls (or ▩ is a gate tile).
                    #
                    # At ▩, the box can also only move along the
                    # axis given by dir.
                    # Therefore, when the box moves in the direction given by
                    # dir to ▩, the box may just as well be pushed one step
                    # further, instead of resting at ▩.
                    #
                    # The following figures depict the situation after the
                    # box has been pushed upwards to ▩;
                    #
                    #  #▩#    #▩_    _▩#    #_▩_#
                    #  #□#    #□#    #□#     #□#
                    #
                    # if ▩ is a gate tile then □ doesn't
                    # need to be blocked by walls on both sides
                    #
                    # additionally, if □ is a gate tile and has
                    # walls on both sides, then "parking" at ▩ is
                    # unnecessary too:
                    #
                    #  _▩_
                    #  #□#    □ is a gate square
                    #
                    if (((game.board[▩] & GOAL) == 0) && # Not a goal
                       (
                        (
                         (tile_is_blocked_by_two_walls_along_another_axis || ((game.board[▩] & FLAG_GATE_TILE) > 0))
                         &&
                         is_box_blocked_along_axis_wrt_walls(game, ▩, NEXT_DIRECTION[dir])
                        )
                        ||
                        (
                         tile_is_blocked_by_two_walls_along_another_axis && ((game.board[□] & FLAG_GATE_TILE) > 0)
                        )
                       ))
                        game.board[▩] |= DIRECTION_TO_TUNNEL_FLAG[dir]
                    end
                end
            end
        end
    end
end