"""
This method determines whether a box placed at the given tile would be blocked
in the given axis. This check only considers walls - not other boxes.
"""
function is_box_blocked_along_axis_wrt_walls(game::Game, □::TileIndex, dir::Direction)::Bool
    v_fore = game.board[□ + game.step_fore[dir]]
    v_rear = game.board[□ - game.step_fore[dir]]
    return (  # if a wall at either neighbor
         ((v_fore & WALL) > 0) ||
         ((v_rear & WALL) > 0) ||
         ( # Both neighbors are illegal box squares
            ((v_fore & FLAG_ILLEGAL_BOX_TILE) > 0) &&
            ((v_rear & FLAG_ILLEGAL_BOX_TILE) > 0)
         )
        ) && is_a_legal_and_box_reachable_tile(game.board[□])
end

"""
This is an internal helper method for is_freezing_move.
It checks to see if the given box tile is blocked along the direction perpendicular
to `dir`. It is a recursive call. It returns two values.
This method uses a ReachableTiles struct in order to mark which
tiles have already passed a freeze test. If the value in `tiles`
is equal to our calc_counter, we are free. If it is greater than our
calc_counter, then we are blocked.
"""
function is_box_blocked_along_one_axis(
    □::TileIndex,
    game::Game,
    s::State,
    dir::Direction, # represents the axis
    freeze_data::ReachableTiles,
    a_box_is_blocked_on_a_non_goal_square::Bool = false,
) :: Tuple{Bool,Bool}
    # Flip the horizontal and vertical directions
    if dir == DIR_UP
        dir = NEXT_DIRECTION[dir]
    else
        dir = DIR_UP
    end

    is_blocked = false
    if (dir == DIR_UP) && (freeze_data.tiles[□] ≥ freeze_data.calc_counter)
        # Use the precomputed value
        is_blocked = freeze_data.tiles[□] > freeze_data.calc_counter
    else
        ▩ = □ - game.step_fore[dir]
        ⬓ = □ + game.step_fore[dir]
        v▩ = s.board[▩]
        v⬓ = s.board[⬓]

        # Temporarily change our target tile to a wall
        s.board[□] |= WALL

        is_wall_on_either_neighbor_square = ((v▩ & WALL) > 0) || ((v⬓ & WALL) > 0)
        are_both_neighbors_illegal = ((v▩ & FLAG_ILLEGAL_BOX_TILE) > 0) &&
                                     ((v⬓ & FLAG_ILLEGAL_BOX_TILE) > 0)
        is_blocked = is_wall_on_either_neighbor_square || are_both_neighbors_illegal

        # Is is insufficient to know whether a box is blocked due to walls -
        # we must also know whether a box is blocked on a non-goal square

        is_▩_blocked = (v▩ & (WALL + BOX)) == BOX
        if (!is_blocked || !a_box_is_blocked_on_a_non_goal_square) && is_▩_blocked
            is_▩_blocked, a_box_is_blocked_on_a_non_goal_square = is_box_blocked_along_one_axis(
                ▩, game, s, dir, freeze_data, a_box_is_blocked_on_a_non_goal_square)
            if is_▩_blocked
                is_blocked = true
            end
        end

        is_⬓_blocked = (v⬓ & (WALL + BOX)) == BOX
        if (!is_blocked || !a_box_is_blocked_on_a_non_goal_square) && is_⬓_blocked
            is_⬓_blocked, a_box_is_blocked_on_a_non_goal_square = is_box_blocked_along_one_axis(
                ⬓, game, s, dir, freeze_data, a_box_is_blocked_on_a_non_goal_square)
            if is_⬓_blocked
                is_blocked = true
            end
        end

        # Remove the wall again
        s.board[□] &= ~ WALL
    end

    # If this box is blocked and is not a goal square
    if is_blocked && ((s.board[□] & GOAL) == 0)
       a_box_is_blocked_on_a_non_goal_square = true
    end

    # Store our result in freeze data
    if dir == DIR_UP
        freeze_data.tiles[□] = freeze_data.calc_counter + is_blocked
    end

    return (is_blocked, a_box_is_blocked_on_a_non_goal_square)
end

"""
Checks whether the given box move creates a frozen game state.
This method requires that box-reachable tiles have been computed.
"""
function is_a_freezing_move(
    □_src::TileIndex, # Can be set to 1 if unused
    □_dst::TileIndex,
    game::Game,
    s::State,
    freeze_data::ReachableTiles,
    a_box_freezing_on_a_goal_square_counts_as_a_freezing_move::Bool,
    )

    # We are certainly frozen if we are a wall or a deadlocked tile,
    # or if a box should not be on this tile to begin with.
    if ((s.board[□_dst] & (WALL + FLAG_ILLEGAL_BOX_TILE)) > 0) ||
       ((s.board[□_dst] & FLAG_BOX_REACHABLE_TILE) == 0)
        return true
    end

    # We have to perform a more complicated check.

    # Remove the box, if any, from the source tile
    v_src_orig = s.board[□_src]
    s.board[□_src] &= ~BOX

    # Increment or reset our freeze data
    if freeze_data.calc_counter < typemax(ReachabilityCount) - 5
        freeze_data.calc_counter += 2
    else
        fill!(freeze_data.tiles, zero(ReachabilityCount))
        freeze_data.calc_counter = 2
    end

    a_box_is_blocked_on_a_non_goal_square = false
    is_blocked_vert, a_box_is_blocked_on_a_non_goal_square = is_box_blocked_along_one_axis(
        □_dst, game, s, DIR_UP, freeze_data, a_box_is_blocked_on_a_non_goal_square)
    is_blocked_horz, a_box_is_blocked_on_a_non_goal_square = is_box_blocked_along_one_axis(
        □_dst, game, s, NEXT_DIRECTION[DIR_UP], freeze_data, a_box_is_blocked_on_a_non_goal_square)

    retval = is_blocked_vert &&
             is_blocked_horz &&
             (a_box_is_blocked_on_a_non_goal_square || a_box_freezing_on_a_goal_square_counts_as_a_freezing_move)

    # Reset the original tile value
    s.board[□_src] = v_src_orig
    return retval
end

"""
The sum of the distance to the nearest goal for all boxes,
which is a lower bound on the number of pushes (and thus also moves) to complete the board.
"""
function calculate_simple_lower_bound(s::State, dist_to_nearest_goal::Vector{Int32})::Int32
    value = zero(Int32)
    for □ in s.boxes
        if □ != 0 # box is on the board / hasn't been temporarily removed
            distance = dist_to_nearest_goal[□]
            if distance == typemax(Int32)
                # This box cannot reachy any goal
                return distance
            end
            value += distance
        end
    end
    return value
end


"""
Get the list of valid pushes for the given state.
NOTE: We assume reachability has already been computed.
NOTE: This method is not all that efficient - it allocates a moves vector.
"""
function get_pushes(game::Game, s::State, reach::ReachableTiles)
    pushes = Push[]
    for (i_box, □) in enumerate(s.boxes)
        if is_reachable_box(□, reach)
            # Get a move for each direction we can get at it from,
            # provided that the opposing side is clear.
            for dir in DIRECTIONS
                □_dest = □ + game.step_fore[dir] # Destination tile
                □_player = □ - game.step_fore[dir] # Side of the player
                # If we can reach the player's pos and the destination is clear.
                if is_reachable(□_player, reach) && ((s.board[□_dest] & (FLOOR+BOX)) == FLOOR)
                    push!(pushes, Push(i_box, dir))
                end
            end
        end
    end
    return pushes
end

"""
Fill pushes with the list of valid pushes for the given state.
We return the number of valid pushes.
NOTE: We assume reachability has already been computed.
"""
function get_pushes!(pushes::Vector{Push}, game::Game, s::State, reach::ReachableTiles)::Int
    n_pushes = 0
    for (i_box, □) in enumerate(s.boxes)
        if is_reachable_box(□, reach)
            # Get a move for each direction we can get at it from,
            # provided that the opposing side is clear.
            for dir in DIRECTIONS
                □_dest = □ + game.step_fore[dir] # Destination tile
                □_player = □ - game.step_fore[dir] # Side of the player
                # If we can reach the player's pos and the destination is clear.
                if is_reachable(□_player, reach) && ((s.board[□_dest] & (FLOOR+BOX)) == FLOOR)
                    n_pushes += 1
                    pushes[n_pushes] = Push(i_box, dir)
                end
            end
        end
    end
    return n_pushes
end

"""
Given a push, produces the next valid push, with pushes ordered
by i_box and then by direction.
The first valid push can be obtained by passing in a box index of 1 and direction 0x00.
If there no valid next push, a push with box index 0 is returned.
"""
function get_next_push!(game::Game, s::State, reach::ReachableTiles, push::Push = Push(1, 0))
    i_box = push.i_box
    dir = push.dir
    while i_box ≤ length(s.boxes)
        □ = s.boxes[i_box]
        if is_reachable_box(□, reach)
            # Get a move for each direction we can get at it from,
            # provided that the opposing side is clear.
            while dir < N_DIRS
                dir += one(Direction)
                □_dest = □ + game.step_fore[dir] # Destination tile
                □_player = □ - game.step_fore[dir] # Side of the player
                # If we can reach the player's pos and the destination is clear.
                if is_reachable(□_player, reach) && ((s.board[□_dest] & (FLOOR+BOX)) == FLOOR)
                    return Push(i_box, dir)
                end
            end
        end

        # Reset
        dir = zero(Direction)
        i_box += one(BoxIndex)
    end

    # No next push
    return Push(0, 0)
end