"""
This solver uses A* via the simple lower bound.
It runs a depth-first search but stops when it reaches a max cost (lowerbound + cost thus far).
Uses a transposition table to avoid expanding states that have already been seen.
"""
struct IterativeDeepeningAstar
    d_max::Int
    cost_limit::Int32
    timeout::Float64 # Maximum execution time in seconds
end
IterativeDeepeningAstar(; d_max::Int=1, cost_limit::Int32=1, timeout::Float64=1.0) = IterativeDeepeningAstar(d_max, cost_limit, timeout)

get_params(solver::IterativeDeepeningAstar) = Dict{String,Any}(
    "d_max"=>solver.d_max,
    "cost_limit"=>solver.cost_limit,
    "timeout"=>solver.timeout,
    )

mutable struct IDAData
    solved::Bool
    max_cost_bound::Int32 # the cost limit when solution was obtained
    max_depth::Int # max depth reached during search
    n_pushes_evaled::Int # number of pushes evaluated during search
    n_pushes_skipped::Int # number of pushes skipped due to the transposition table
    n_infeasible_states::Int
    setup_time::Float64 # time it takes to set up the problem
    total_time::Float64
    pushes::Vector{Push} # of length d_max
    sol_depth::Int # number of pushes in the solution
end

function print_results(data::IDAData, verbosity::Int=1)
    println("\t\tsolved: ", data.solved)
    println("\t\ttotal time: ", data.total_time)
    if verbosity > 0
        println("\t\tsetup_time: ", data.setup_time, " ($(round(data.setup_time/data.total_time*100,digits=2))%)")
        println("\t\tsolving time: ", data.total_time - data.setup_time)
        println("\t\tmax_cost_bound: ", data.max_cost_bound)
        println("\t\tmax_depth: ", data.max_depth)
        println("\t\tsol_depth: ", data.sol_depth)
        println("\t\tn_pushes_evaled: ", data.n_pushes_evaled)
        println("\t\tn_pushes_skipped: ", data.n_pushes_skipped)
        println("\t\tn_infeasible_states: ", data.n_infeasible_states)
    end
end

function solve_(
    solver::IterativeDeepeningAstar,
    data::IDAData,
    game::Game,
    s::State,
    d::Int, # push depth
    lowerbound::Int32, # Current lowerbound value
    cost_bound::Int32, # Disregard states with lowerbounds greater than this
    stack::Vector{TileIndex},
    reach::ReachableTiles,
    pushes_per_depth::Vector{Vector{Push}},
    dist_to_nearest_goal::Vector{Int32},
    transposition_table::Dict{Tuple{TileIndex, UInt64},Int}, # (player normalized pos, box zhash) → d_min
    )

    data.max_depth = max(data.max_depth, d)

    if lowerbound == 0
        # Solved!
        data.solved = true
        data.sol_depth = d
        return true
    end

    if d + lowerbound > cost_bound
        # We exceeded our cost bound
        return false
    end

    if d ≥ solver.d_max
        # We exceeded our max depth
        return false
    end

    # Calc reachability for this state
    calc_reachable_tiles!(reach, game, s.board, s.player, stack)

    # Set the position to the normalized location
    s.player = reach.min_reachable_tile

    # Exit early if we have seen this state before at this shallow of a depth
    ztup = (s.player, s.zhash)
    d_seen = get(transposition_table, ztup, typemax(Int))
    if d_seen ≤ d
        # Skip!
        data.n_pushes_skipped += 1
        return false
    end

    # Store this state hash in our transposition table
    transposition_table[ztup] = d

    pushes = pushes_per_depth[d+1]
    n_pushes = get_pushes!(pushes, game, s, reach)
    for push_index in 1:n_pushes
        push = pushes[push_index]
        data.n_pushes_evaled += 1

        # Move, with an incremental update of the lowerbound
        # NOTE: We are guaranteed to only push a box with a finite dist-to-nearest-goal
        lowerbound_new = lowerbound - dist_to_nearest_goal[s.boxes[push.i_box]]
        move!(s, game, push)
        lowerbound_component = dist_to_nearest_goal[s.boxes[push.i_box]]
        if lowerbound_component == typemax(Int32)
            # We exceeded!
            # Do not proceed
            data.n_infeasible_states += 1
        else
            lowerbound_new += lowerbound_component
            if solve_(solver, data, game, s, d+1, lowerbound_new, cost_bound, stack, reach, pushes_per_depth, dist_to_nearest_goal, transposition_table)
                # Store the solution
                data.pushes[d+1] = push
                return true
            end
        end

        unmove!(s, game, push)
    end

    # Failed to find a solution
    return false
end

function solve(solver::IterativeDeepeningAstar, game::Game, s::State)
    t_start = time()

    #########################################################################################
    # Allocate memory and recalculate some things
    d_max, cost_limit, timeout = solver.d_max, solver.cost_limit, solver.timeout

    reach = ReachableTiles(length(game.board))
    clear!(reach, game.board)
    stack = Array{TileIndex}(undef, length(game.board))
    pushes_per_depth = Array{Vector{Push}}(undef, d_max)
    for d in 1:d_max
        pushes_per_depth[d] = Array{Push}(undef, 4*length(s.boxes))
    end

    # Calculate reachability and set player min pos
    calc_reachable_tiles!(reach, game, s.board, s.player, stack)
    move_player!(s, reach.min_reachable_tile)

    # Calculate distance to nearest goal for all squares
    queue_big = Array{Tuple{TileIndex, TileIndex, Int32}}(undef, length(s.board) * N_DIRS)
    distances_by_direction = Array{Int32}(undef, length(s.board), N_DIRS)
    dist_to_nearest_goal = zeros(Int32, length(s.board))
    calc_push_dist_to_nearest_goal_for_all_squares!(dist_to_nearest_goal, game, s, reach, stack, queue_big, distances_by_direction)

    # Transposition table
    transposition_table = Dict{Tuple{TileIndex, UInt64}, Int}()

    # Compute the lower bound
    lowerbound = calculate_simple_lower_bound(s, dist_to_nearest_goal)

    # Get our data ready
    data = IDAData(false, zero(Int32), 0, 0, 0, 0, 0.0, 0.0, Array{Push}(undef, d_max), 0)

    data.setup_time = time() - t_start
    #########################################################################################
    # Run solver

    # Start at our current lowerbound, otherwise we wouldn't expand at all
    cost_bound = lowerbound
    while !data.solved && cost_bound < cost_limit && (time() - t_start < timeout)
        empty!(transposition_table)
        solve_(solver, data, game, s, 0, lowerbound, cost_bound, stack, reach, pushes_per_depth, dist_to_nearest_goal, transposition_table)
        cost_bound += one(Int32)
    end

    data.max_cost_bound = cost_bound - one(Int32)
    data.total_time = time() - t_start
    return data
end