#=
A depth-first search implementation of packing order search, similar to how YASS operates.
http://sokobano.de/wiki/index.php?title=Sokoban_solver_%22scribbles%22_by_Brian_Damgaard_about_the_YASS_solver#Goal_room_packing

Start State: same as original problem
Goal State: same as original problem - all boxes on goals

State:
- The same as a Game state, except we additionally track, for each box, whether it is
    - not moved
    - temporarily parked
    - permanently parked

Action:
- A box that has never been moved can either be temporarily parked or placed on a goal,
  assuming said destination is reachable by pushes
- A box that has been temporarily parked can be placed on a goal,
  assuming said destination is reachable by pushes
- Valid parking tiles are either:
    - the farthest in manhattan distance from its start
    - or the farthest in push distance from its start

Algorithm:
- DFS for starters
  - if there is no re-parking of parked boxes, then max depth is 2*|b| for |b| boxes
  - branching factor is pretty big, at |b| + 2 for the first action, as we can push to any of the |b| available start squares and 2 parking tiles, etc.
=#

struct PackingOrderSearch
    timeout::Float64 # Maximum execution time in seconds
end

get_params(solver::PackingOrderSearch) = Dict{String,Any}(
    "timeout"=>solver.timeout,
)

struct PackingOrderSearchAction
    bring_home::Bool      # If true, we are pushing the box to its final destination. If false, we are temporarily parking it at its destination
    i_box::BoxIndex       # The box we are pulling
    □::TileIndex          # The destination tile (either a final destination or a parking tile)
    dir::Direction        # The direction the player ends up at from the box after the push
end

const BOX_STATUS_UNMOVED            = 0x00
const BOX_STATUS_TEMPORARILY_PARKED = 0x01
const BOX_STATUS_PLACED             = 0x02

struct PackingOrderSearchMemory
    reach::ReachableTiles
    stack::Vector{TileIndex} # Used for reachability calcs
    queue::Vector{Tuple{TileIndex, TileIndex, Int32}} # Of length board_size * N_DIRS, (box, player, distance)
    distances_by_direction::Matrix{Int32} # Of size board_size × N_DIRS
    box_statuses::Vector{UInt8} # For each box, its status (see enum above)
    t_start::Float64 # The start time
end

mutable struct PackingOrderSearchData
    solved::Bool
    max_depth::Int                            # max depth reached during search
    n_macro_actions_evaled::Int               # total number of macro actions evaluated
    n_parking_tile_actions_evaled::Int        # number of macro actions evaluated that were temporarily parking a box
    setup_time::Float64                       # time it takes to set up the problem
    total_time::Float64
    actions::Vector{PackingOrderSearchAction} # preallocated, of length 2*|b|
    sol_depth::Int                            # number of actions in the solution
end

function print_results(data::PackingOrderSearchData, verbosity::Int=1)
    println("\t\tsolved: ", data.solved)
    println("\t\ttotal time: ", data.total_time)
    if verbosity > 0
        println("\t\tsetup_time: ", data.setup_time, " ($(round(data.setup_time/data.total_time*100,digits=2))%)")
        println("\t\tsolving time: ", data.total_time - data.setup_time)
        println("\t\tmax_depth: ", data.max_depth)
        println("\t\tsol_depth: ", data.sol_depth)
        println("\t\tn_macro_actions_evaled: ", data.n_macro_actions_evaled)
        println("\t\tn_parking_tile_actions_evaled: ", data.n_parking_tile_actions_evaled)
    end
end

# Calculate our actions
# For each unplaced box
#    See where we can pull it
#    Any goals are potential destinations
#    If box is unmoved, try the two temporary goal squares
function get_macro_actions(game::Game, s::State, mem::PackingOrderSearchMemory)

    actions = PackingOrderSearchAction[]

    for i_box in box_indices(game)
        # If the box is in its final resting spot, skip it
        if mem.box_statuses[i_box] == BOX_STATUS_PLACED
            continue
        end

        □_box = s.boxes[i_box]
        calc_box_push_distances!(
            mem.distances_by_direction, game, s, □_box,
            mem.reach, mem.stack, mem.queue)

        # See if we can reach any goals
        for i_goal in goal_indices(game)
            □_goal = game.goals[i_goal]
            for dir in DIRECTIONS
                if !is_type_extrema(mem.distances_by_direction[□_goal, dir])
                    # The box can be pushed to this position
                    push!(actions, PackingOrderSearchAction(
                        true, # bring it home
                        i_box,
                        □_goal,
                        dir
                    ))
                end
            end
        end

        # If the box has not been moved yet, find the two candidate parking tiles
        if mem.box_statuses[i_box] == BOX_STATUS_UNMOVED
            # Find a farthest manhattan distance tile and a farthest push distance tile
            □_manhattan = □_box
            □_push = □_box
            dir_manhattan = DIR_UP
            dir_push = DIR_UP
            d_manhattan = 0
            d_push = 0
            col_start, row_start = tile_index_to_col_row(game, □_box)
            for □ in tile_indices(game)
                if (game.board[□] & GOAL) == 0 # not a goal - we check those

                    # get the min-push distance to this tile
                    d_tile = typemax(eltype(mem.distances_by_direction))
                    fastest_dir = DIR_UP
                    for dir in DIRECTIONS
                        if !is_type_extrema(mem.distances_by_direction[□, dir])
                            d_tile = min(d_tile, mem.distances_by_direction[□, dir])
                            fastest_dir = dir
                        end
                    end

                    if d_tile != typemax(eltype(mem.distances_by_direction))
                        # The box can be pushed to this (non-goal) position
                        col, row = tile_index_to_col_row(game, □)

                        d_manhattan_candidate = abs(col-col_start) + abs(row - row_start)
                        if d_manhattan_candidate > d_manhattan
                            (□_manhattan, dir_manhattan, d_manhattan) = (□, fastest_dir, d_manhattan_candidate)
                        end

                        d_push_candidate = d_tile
                        if d_push_candidate > d_push || (d_push_candidate == d_push && □ != □_manhattan)
                            (□_push, dir_push, d_push) = (□, fastest_dir, d_push_candidate)
                        end
                    end
                end
            end

            if □_manhattan != □_box
                # Add a stashing move for this max-manhattan distance tile
                push!(actions, PackingOrderSearchAction(
                        false, # temporarily place it on this tile
                        i_box,
                        □_manhattan,
                        dir_manhattan
                    ))
            end

            if □_push != □_box && (□_push != □_manhattan || dir_push != dir_manhattan)
                # Add a stashing move for this max-push distance tile
                push!(actions, PackingOrderSearchAction(
                        false, # temporarily place it on this tile
                        i_box,
                        □_push,
                        dir_push
                    ))
            end
        end
    end

    return actions
end

# Move according to a macro action, modifying our state
function move!(s::State, box_statuses::Vector{UInt8}, game::Game, a::PackingOrderSearchAction)
    # Update our box status
    box_statuses[a.i_box] = a.bring_home ? BOX_STATUS_PLACED : BOX_STATUS_TEMPORARILY_PARKED

    # Move the box and player
    □ = s.boxes[a.i_box] # where box starts
    ▩ = a.□ # where box ends up

    s.board[s.player] &= ~PLAYER # Clear the player
    s.board[□] &= ~BOX # Clear the box
    s.zhash ⊻= game.hash_components[□] # Undo the box's hash component

    s.boxes[a.i_box] = ▩ # Place the box
    s.player =  ▩ + game.step_fore[a.dir] # Player ends up next to the box

    s.board[s.player] |= PLAYER # Add the player
    s.board[▩] |= BOX # Add the box
    s.zhash ⊻= game.hash_components[▩] # Add the box's hash component

    return s
end

# Undo a macro move, modifying our state
function unmove!(
    s::State,
    box_statuses::Vector{UInt8},
    game::Game,
    a::PackingOrderSearchAction,
    □_box_prev::TileIndex,
    □_player_prev::TileIndex,
    box_status_prev::UInt8)

    # Revert our box status
    box_statuses[a.i_box] = box_status_prev

    # Move the box and player
    □ = s.boxes[a.i_box] # where box starts
    ▩ = □_box_prev # where box ends up

    s.board[s.player] &= ~PLAYER # Clear the player
    s.board[□] &= ~BOX # Clear the box
    s.zhash ⊻= game.hash_components[□] # Undo the box's hash component

    s.player = □_player_prev
    s.boxes[a.i_box] = ▩

    s.board[s.player] |= PLAYER # Add the player
    s.board[▩] |= BOX # Add the box
    s.zhash ⊻= game.hash_components[▩] # Add the box's hash component

    return s
end

function solve_(
    solver::PackingOrderSearch,
    mem::PackingOrderSearchMemory,
    data::PackingOrderSearchData,
    game::Game,
    s::State,
    d::Int)

    data.max_depth = max(data.max_depth, d)

    # Solved if all boxes are at final destinations
    if all(status == BOX_STATUS_PLACED for status in mem.box_statuses)
        # Solved!
        data.solved = true
        data.sol_depth = d
        return true
    end

    if time() - mem.t_start > solver.timeout
        # Exceeded our time
        return false
    end

    actions = get_macro_actions(game, s, mem)
    for action in actions
        data.n_macro_actions_evaled += 1
        □_box_prev = s.boxes[action.i_box]
        □_player_prev = s.player
        box_status_prev = mem.box_statuses[action.i_box]
        move!(s, mem.box_statuses, game, action)
        if solve_(solver, mem, data, game, s, d+1)
            # Store the solution
            data.actions[d+1] = action
            return true
        end
        unmove!(s, mem.box_statuses, game, action, □_box_prev, □_player_prev, box_status_prev)
    end

    # Failed to find a solution
    return false
end

function solve(solver::PackingOrderSearch, game::Game, s::State)
    t_start = time()

    n_tiles = length(game.board)
    n_boxes = length(game.goals)

    mem = PackingOrderSearchMemory(
        ReachableTiles(n_tiles),
        Array{TileIndex}(undef, n_tiles),
        Array{Tuple{TileIndex, TileIndex, Int32}}(undef, n_tiles * N_DIRS),
        Array{Int32}(undef, (n_tiles, N_DIRS)),
        zeros(n_boxes),
        t_start
    )

    d_max = 2*n_boxes
    data = PackingOrderSearchData(false, 0, 0, 0, 0.0, 0.0, Array{PackingOrderSearchAction}(undef, d_max), 0)

    clear!(mem.reach, s.board)

    solve_(solver, mem, data, game, s, 0)
    data.total_time = time() - t_start
    return data
end

"""
Convert a packing order search action to the minimum sequence of pushes
"""
function convert_to_pushes(
    game::Game,
    s::State,
    a::PackingOrderSearchAction,
    mem::PackingOrderSearchMemory,
    previous_states::Matrix{Tuple{TileIndex, TileIndex}} # Of size board_size × N_DIRS, (box index, player index)
    )

    □_box = s.boxes[a.i_box]
    calc_box_push_distances!(
            mem.distances_by_direction, previous_states, game, s, □_box,
            mem.reach, mem.stack, mem.queue)

    # Box destination
    ▩_box = a.□

    # Figure out the best reachable initial direction of the player from the box
    last_dir = a.dir
    □_player = ▩_box + game.step_fore[last_dir]
    n_pushes_to_append = mem.distances_by_direction[▩_box, last_dir]
    @assert 0 ≤ n_pushes_to_append < 100 # sanity check
    calc_reachable_tiles!(mem.reach, game, s.board, □_player, mem.stack)
    for dir in DIRECTIONS
        if is_reachable(▩_box + game.step_fore[dir], mem.reach) &&
           mem.distances_by_direction[▩_box, dir] < n_pushes_to_append
            last_dir = dir
            n_pushes_to_append = mem.distances_by_direction[▩_box, dir]
        end
    end

    # We know the push distance, so we can append all necessary pushes and then fill them in
    pushes = Array{Push}(undef, n_pushes_to_append)

    # Start at the destination and work our way back
    for i_push in n_pushes_to_append:-1:1
        □_box_prev, □_player_prev = previous_states[▩_box, last_dir]
        @assert □_box_prev != 0
        @assert □_player_prev != 0

        # Push direction is based on where the player pushed from
        # Last dir gets updated to the direction the player was from the box
        push_dir = typemax(Direction)
        for dir in DIRECTIONS
            if ▩_box == □_box_prev + game.step_fore[dir]
                push_dir = dir
            end
            if □_box_prev + game.step_fore[dir] == □_player_prev
                last_dir = dir
            end
        end
        @assert push_dir != typemax(Direction)

        pushes[i_push] = Push(a.i_box, push_dir)
        ▩_box = □_box_prev
    end

    return pushes
end

"""
Convert a trajectory in packing order search actions from the given start state to a trajectory in pushes.
The given state should have the player in their correct starting position.
NOTE: We allocate memory.
"""
function convert_to_pushes(game::Game, s0::State, actions::Vector{PackingOrderSearchAction})
    # Allocate memory
    s = deepcopy(s0)

    n_tiles = length(game.board)
    n_boxes = length(game.goals)

    mem = PackingOrderSearchMemory(
        ReachableTiles(n_tiles),
        Array{TileIndex}(undef, n_tiles),
        Array{Tuple{TileIndex, TileIndex, Int32}}(undef, n_tiles * N_DIRS),
        Array{Int32}(undef, (n_tiles, N_DIRS)),
        zeros(n_boxes),
        time()
    )

    clear!(mem.reach, s.board)
    previous_states = Matrix{Tuple{TileIndex, TileIndex}}(undef, (n_tiles, N_DIRS))

    pushes = Push[]
    for a in actions
        # Convert the action into pushes
        append!(pushes, convert_to_pushes(game, s, a, mem, previous_states))

        # Now apply the action
        move!(s, mem.box_statuses, game, a)
    end

    return pushes
end