#=
A depth-first search implementation of packing order search, similar to how YASS operates.
The "rev" refers to "reverse", where this version of packing order search, like YASS, does pulls
from a goal state to the initial state.
http://sokobano.de/wiki/index.php?title=Sokoban_solver_%22scribbles%22_by_Brian_Damgaard_about_the_YASS_solver#Goal_room_packing

Note that a revered solution can be better for sokoban problems, because it is supposed
to reduce the tendency to end up in deadlock situations. The idea is that the solution is
low-entropy, and we can maintain relatively low entropy with our pulls, but it is harder
to avoid deadlocks by pushing from the high-entropy start state.

Start State: all boxes on goals
  Note that we do not know where the player ends up, so we have to pick an open spot next to a box and start there.
  There may be multiple such spots.
Goal State: the start state of the orginal problem

State:
- The same as a Game state, except we additionally track, for each box, whether it is
    - not moved
    - temporarily parked
    - permanently parked

Action:
- A box that has never been moved can either be temporarily parked or placed on a goal,
  assuming said destination is reachable by pulls
- A box that has been temporarily parked can be placed on a goal,
  assuming said destination is reachable by pulls
- Valid parking tiles are either:
    - the farthest in manhattan distance from its start
    - or the farthest in pull distance from its start
=#

struct PackingOrderSearchRev
    timeout::Float64 # Maximum execution time in seconds
end

get_params(solver::PackingOrderSearchRev) = Dict{String,Any}(
    "timeout"=>solver.timeout,
)

struct PackingOrderSearchRevAction
    bring_home::Bool      # If true, we are pulling the box to its final destination. If false, we are temporarily parking it at its destination
    i_box::BoxIndex       # The box we are pulling
    □::TileIndex          # The destination tile (either a final destination or a parking tile)
    dir::Direction        # The direction the player ends up at from the box after the pull
end

const BOX_STATUS_UNMOVED            = 0x00
const BOX_STATUS_TEMPORARILY_PARKED = 0x01
const BOX_STATUS_PLACED             = 0x02

struct PackingOrderSearchRevMemory
    reach::ReachableTiles
    stack::Vector{TileIndex} # Used for reachability calcs
    queue::Vector{Tuple{TileIndex, TileIndex, Int32}} # Of length board_size * N_DIRS, (box, player, distance)
    distances_by_direction::Matrix{Int32} # Of size board_size × N_DIRS
    box_statuses::Vector{UInt8} # For each box, its status (see enum above)
    t_start::Float64  # The start time
end

mutable struct PackingOrderSearchRevData
    solved::Bool
    max_depth::Int                            # max depth reached during search
    n_macro_actions_evaled::Int               # total number of macro actions evaluated
    n_parking_tile_actions_evaled::Int        # number of macro actions evaluated that were temporarily parking a box
    setup_time::Float64                       # time it takes to set up the problem
    total_time::Float64
    actions::Vector{PackingOrderSearchRevAction} # preallocated, of length 2*|b|
    sol_depth::Int                            # number of actions in the solution
    end_player_tile::TileIndex                # tile index player ends at (i.e. our start position in reverse-mode)
end

function print_results(data::PackingOrderSearchRevData, verbosity::Int=1)
    println("\t\tsolved: ", data.solved)
    println("\t\ttotal time: ", data.total_time)
    if verbosity > 0
        println("\t\tsetup_time: ", data.setup_time, " ($(round(data.setup_time/data.total_time*100,digits=2))%)")
        println("\t\tsolving time: ", data.total_time - data.setup_time)
        println("\t\tmax_depth: ", data.max_depth)
        println("\t\tsol_depth: ", data.sol_depth)
        println("\t\tn_macro_actions_evaled: ", data.n_macro_actions_evaled)
        println("\t\tn_parking_tile_actions_evaled: ", data.n_parking_tile_actions_evaled)
        println("\t\tend_player_tile: ", data.end_player_tile)
    end
end

# Calculate our start states
# The start state has all boxes on goals
# We then compute all player-reachable regions, and return
# the min-index player reachable tile for each.
function calc_reverse_mode_start_states(game::Game, reach::ReachableTiles, stack::Vector{TileIndex})

    # Construct a state object that copies the current board
    s = State(game, deepcopy(game.board))

    # Remove the player and all boxes
    move_player!(s, zero(TileIndex))
    remove_all_boxes_from_board!(s)

    # Place boxes on all goals
    copy!(s.boxes, game.goals)
    place_all_boxes_on_board!(s)


    # Compute all player-reachable regions (and thus potential end states)
    □_players = TileIndex[]
    clear!(reach, s.board) # floor tiles set to zero, walls set to typemax

    while true
        # Find a non-goal floor tile we have not reached yet
        # (These tiles will have a reach counter of zero)
        □_start = zero(TileIndex)
        for □ in tile_indices(game)
            if reach.tiles[□] == 0 && (s.board[□] & GOAL) == 0
                □_start = □
                break
            end
        end

        if □_start == 0
            break
        end

        # Calculate reachability
        calc_reachable_tiles!(reach, game, s.board, TileIndex(□_start), stack)
        push!(□_players, reach.min_reachable_tile)
    end

    return (s, □_players)
end

# Calculate our actions
# For each unplaced box
#    See where we can pull it
#    Any box starting positions are potential destinations
#    If box is unmoved, try the two temporary goal squares
function get_macro_actions(game::Game, s::State, mem::PackingOrderSearchRevMemory)

    actions = PackingOrderSearchRevAction[]

    for i_box in box_indices(game)
        # If the box is in its final resting spot, skip it
        if mem.box_statuses[i_box] == BOX_STATUS_PLACED
            continue
        end

        □_box = s.boxes[i_box]
        calc_box_pull_distances!(
            mem.distances_by_direction, game, s, □_box,
            mem.reach, mem.stack, mem.queue)

        # See if we can reach any box starting positions
        for i_goal in goal_indices(game)
            □_goal = game.starting_boxes[i_goal]
            for dir in DIRECTIONS
                if !is_type_extrema(mem.distances_by_direction[□_goal, dir])
                    # The box can be pulled to this position
                    push!(actions, PackingOrderSearchRevAction(
                        true, # bring it home
                        i_box,
                        □_goal,
                        dir
                    ))
                end
            end
        end

        # If the box has not been moved yet, find the two candidate parking tiles
        if mem.box_statuses[i_box] == BOX_STATUS_UNMOVED
            # Find the farthest manhattan distance tile and the farthest pull distance tile
            □_manhattan = □_box
            □_pull = □_box
            dir_manhattan = DIR_UP
            dir_pull = DIR_UP
            d_manhattan = 0
            d_pull = 0
            col_start, row_start = tile_index_to_col_row(game, □_box)
            for □ in tile_indices(game)
                if (game.board[□] & GOAL) == 0 # not a goal - we check those

                    # get the min-pull distance to this tile
                    d_tile = typemax(eltype(mem.distances_by_direction))
                    fastest_dir = DIR_UP
                    for dir in DIRECTIONS
                        if !is_type_extrema(mem.distances_by_direction[□, dir])
                            d_tile = min(d_tile, mem.distances_by_direction[□, dir])
                            fastest_dir = dir
                        end
                    end

                    if d_tile != typemax(eltype(mem.distances_by_direction))
                        # The box can be pulled to this (non-goal) position
                        col, row = tile_index_to_col_row(game, □)

                        d_manhattan_candidate = abs(col-col_start) + abs(row - row_start)
                        if d_manhattan_candidate > d_manhattan
                            (□_manhattan, dir_manhattan, d_manhattan) = (□, fastest_dir, d_manhattan_candidate)
                        end

                        d_pull_candidate = d_tile
                        if d_pull_candidate > d_pull || (d_pull_candidate == d_pull && □ != □_manhattan)
                            (□_pull, dir_pull, d_pull) = (□, fastest_dir, d_pull_candidate)
                        end
                    end
                end
            end

            if □_manhattan != □_box
                # Add a stashing move for this max-manhattan distance tile
                push!(actions, PackingOrderSearchRevAction(
                        false, # temporarily place it on this tile
                        i_box,
                        □_manhattan,
                        dir_manhattan
                    ))
            end

            if □_pull != □_box && (□_pull != □_manhattan || dir_pull != dir_manhattan)
                # Add a stashing move for this max-pull distance tile
                push!(actions, PackingOrderSearchRevAction(
                        false, # temporarily place it on this tile
                        i_box,
                        □_pull,
                        dir_pull
                    ))
            end
        end
    end

    return actions
end

# Move according to a macro action, modifying our state
function move!(s::State, box_statuses::Vector{UInt8}, game::Game, a::PackingOrderSearchRevAction)
    # Update our box status
    box_statuses[a.i_box] = a.bring_home ? BOX_STATUS_PLACED : BOX_STATUS_TEMPORARILY_PARKED

    # Move the box and player
    □ = s.boxes[a.i_box] # where box starts
    ▩ = a.□ # where box ends up

    s.board[s.player] &= ~PLAYER # Clear the player
    s.board[□] &= ~BOX # Clear the box
    s.zhash ⊻= game.hash_components[□] # Undo the box's hash component

    s.boxes[a.i_box] = ▩ # Place the box
    s.player =  ▩ + game.step_fore[a.dir] # Player ends up next to the box

    s.board[s.player] |= PLAYER # Add the player
    s.board[▩] |= BOX # Add the box
    s.zhash ⊻= game.hash_components[▩] # Add the box's hash component

    return s
end

# Undo a macro move, modifying our state
function unmove!(
    s::State,
    box_statuses::Vector{UInt8},
    game::Game,
    a::PackingOrderSearchRevAction,
    □_box_prev::TileIndex,
    □_player_prev::TileIndex,
    box_status_prev::UInt8)

    # Revert our box status
    box_statuses[a.i_box] = box_status_prev

    # Move the box and player
    □ = s.boxes[a.i_box] # where box starts
    ▩ = □_box_prev # where box ends up

    s.board[s.player] &= ~PLAYER # Clear the player
    s.board[□] &= ~BOX # Clear the box
    s.zhash ⊻= game.hash_components[□] # Undo the box's hash component

    s.player = □_player_prev
    s.boxes[a.i_box] = ▩

    s.board[s.player] |= PLAYER # Add the player
    s.board[▩] |= BOX # Add the box
    s.zhash ⊻= game.hash_components[▩] # Add the box's hash component

    return s
end

function solve_(
    solver::PackingOrderSearchRev,
    mem::PackingOrderSearchRevMemory,
    data::PackingOrderSearchRevData,
    game::Game,
    s::State,
    d::Int)

    data.max_depth = max(data.max_depth, d)

    # Solved if all boxes are at final destinations
    # and the player is in the same reachable region as the start player tile
    if all(status == BOX_STATUS_PLACED for status in mem.box_statuses)
        # Ensure we can reach the start tile
        calc_reachable_tiles!(mem.reach, game, s.board, s.player, mem.stack)
        if is_reachable(game.start_player_tile, mem.reach)
            data.solved = true
            data.sol_depth = d
            return true
        else
            return false
        end
    end

    if time() - mem.t_start > solver.timeout
        # Exceeded our time
        return false
    end

    actions = get_macro_actions(game, s, mem)
    for action in actions
        data.n_macro_actions_evaled += 1
        □_box_prev = s.boxes[action.i_box]
        □_player_prev = s.player
        box_status_prev = mem.box_statuses[action.i_box]
        move!(s, mem.box_statuses, game, action)
        if solve_(solver, mem, data, game, s, d+1)
            # Store the solution
            data.actions[d+1] = action
            return true
        end
        unmove!(s, mem.box_statuses, game, action, □_box_prev, □_player_prev, box_status_prev)
    end

    # Failed to find a solution
    return false
end

function solve(solver::PackingOrderSearchRev, game::Game)
    t_start = time()

    n_tiles = length(game.board)
    n_boxes = length(game.goals)

    mem = PackingOrderSearchRevMemory(
        ReachableTiles(n_tiles),
        Array{TileIndex}(undef, n_tiles),
        Array{Tuple{TileIndex, TileIndex, Int32}}(undef, n_tiles * N_DIRS),
        Array{Int32}(undef, (n_tiles, N_DIRS)),
        zeros(n_boxes),
        t_start
    )

    d_max = 2*n_boxes
    data = PackingOrderSearchRevData(false, 0, 0, 0, 0.0, 0.0, Array{PackingOrderSearchRevAction}(undef, d_max), 0, 0)

    # Construct the start states
    s, player_start_tiles = calc_reverse_mode_start_states(game, mem.reach, mem.stack)
    clear!(mem.reach, s.board)

    # Compute illegal box squares
    dist_to_nearest_goal = zeros(Int32, n_tiles)
    calc_push_dist_to_nearest_goal_for_all_squares!(dist_to_nearest_goal, game, s, mem.reach, mem.stack, mem.queue, mem.distances_by_direction)
    set_illegal_box_tiles!(game, dist_to_nearest_goal)

    # Run solve for every player start tile, until we find a solution
    # (Just return on the first success)
    for □_player in player_start_tiles
        move_player!(s, □_player)
        solved = solve_(solver, mem, data, game, s, 0)
        if solved
            data.end_player_tile = □_player
            break
        end
    end

    data.total_time = time() - t_start
    return data
end

"""
Convert a PackingOrderSearchRevAction to the minimum sequence of pulls
"""
function convert_to_pulls(
    game::Game,
    s::State,
    a::PackingOrderSearchRevAction,
    mem::PackingOrderSearchRevMemory,
    previous_states::Matrix{Tuple{TileIndex, TileIndex}} # Of size board_size × N_DIRS, (box index, player index)
    )

    □_box = s.boxes[a.i_box]
    calc_box_pull_distances!(
            mem.distances_by_direction, previous_states, game, s, □_box,
            mem.reach, mem.stack, mem.queue)

    # Box destination
    ▩_box = a.□

    # Figure out the best reachable initial direction of the player from the box
    last_dir = a.dir
    □_player = ▩_box + game.step_fore[last_dir]
    n_pulls_to_append = mem.distances_by_direction[▩_box, last_dir]
    @assert 0 ≤ n_pulls_to_append < 100 # sanity check
    calc_reachable_tiles!(mem.reach, game, s.board, □_player, mem.stack)
    for dir in DIRECTIONS
        if is_reachable(▩_box + game.step_fore[dir], mem.reach) &&
           mem.distances_by_direction[▩_box, dir] < n_pulls_to_append
            last_dir = dir
            n_pulls_to_append = mem.distances_by_direction[▩_box, dir]
        end
    end

    # We know the pull distance, so we can append all necessary pulls and then fill them in
    pulls = Array{Push}(undef, n_pulls_to_append)

    # Start at the destination and work our way back
    for i_pull in n_pulls_to_append:-1:1
        □_box_prev, □_player_prev = previous_states[▩_box, last_dir]

        @assert □_box_prev != 0
        @assert □_player_prev != 0

        # Pull direction is based on where the player pulled from
        # Last dir gets updated to the direction the player was from the box
        pull_dir = typemax(Direction)
        for dir in DIRECTIONS
            if ▩_box == □_box_prev + game.step_fore[dir]
                pull_dir = dir # direction of the player from the box
            end
            if □_box_prev + game.step_fore[dir] == □_player_prev
                last_dir = dir
            end
        end
        @assert pull_dir != typemax(Direction)

        pulls[i_pull] = Push(a.i_box, pull_dir) # NOTE: This is a pull
        ▩_box = □_box_prev
    end

    return pulls
end

"""
Convert a trajectory in PackingOrderSearchRevActions from the given start state to a
trajectory in pushes. The given state should have the player in their correct starting position.
NOTE: We allocate memory.
"""
function convert_to_pulls(game::Game, end_player_tile::TileIndex, actions::Vector{PackingOrderSearchRevAction})
    n_tiles = length(game.board)
    n_boxes = length(game.goals)

    mem = PackingOrderSearchRevMemory(
        ReachableTiles(n_tiles),
        Array{TileIndex}(undef, n_tiles),
        Array{Tuple{TileIndex, TileIndex, Int32}}(undef, n_tiles * N_DIRS),
        Array{Int32}(undef, (n_tiles, N_DIRS)),
        zeros(n_boxes),
        time()
    )

    clear!(mem.reach, game.board)
    previous_states = Matrix{Tuple{TileIndex, TileIndex}}(undef, (n_tiles, N_DIRS))

    # Construct the "start" state
    s, player_start_tiles = calc_reverse_mode_start_states(game, mem.reach, mem.stack)
    move_player!(s, data.end_player_tile)
    clear!(mem.reach, s.board)

    # Compute illegal box squares
    dist_to_nearest_goal = zeros(Int32, n_tiles)
    calc_push_dist_to_nearest_goal_for_all_squares!(dist_to_nearest_goal, game, s, mem.reach, mem.stack, mem.queue, mem.distances_by_direction)
    set_illegal_box_tiles!(game, dist_to_nearest_goal)

    pulls = Push[]
    for a in actions
        # Convert the action into pulls
        append!(pulls, convert_to_pulls(game, s, a, mem, previous_states))

        # Now apply the action
        move!(s, mem.box_statuses, game, a)
    end

    return pulls
end

function convert_to_pushes(game::Game, end_player_tile::TileIndex, actions::Vector{PackingOrderSearchRevAction})
    # Compute the pulls
    pulls = convert_to_pulls(game, data.end_player_tile, data.actions[1:data.sol_depth])

    # Get the start board by applying the macro actions in reverse
    # Otherwise our box indices don't match
    s, player_start_tiles = calc_reverse_mode_start_states(game, mem.reach, mem.stack)
    move_player!(s, data.end_player_tile)
    for a in data.actions[1:data.sol_depth]
        move!(s, mem.box_statuses, game, a)
    end

    # Figure out what the correct box indices should be
    i_box_to_box_index = Array{BoxIndex}(undef, length(game.starting_boxes))
    for i_box in box_indices(game)
        for j_box in box_indices(game)
            if s.boxes[j_box] == game.starting_boxes[i_box]
                i_box_to_box_index[i_box] = j_box
            end
        end
    end

    # Now map boxes to their start state indices and reverse the pulls
    pushes = [Push(i_box_to_box_index[p.i_box], OPPOSITE_DIRECTION[p.dir]) for p in reverse(pulls)]
    return pushes
end